/* isc_init.c */

#include <tk.h>
#include <stdlib.h>  /* getenv() */
#include <sys/types.h>
#include <stdio.h>
#include <sys/socket.h>      /* struct sockaddr */
#include <netdb.h>           /* gethostbyname() */
#include <netinet/in.h>      /* sockaddr_in */
#include <arpa/inet.h>       /* inet_addr() */             
#include <tcl.h>
#include <netinet/in.h>

#define MULTICAST(x)       (((x) & htonl(0xf0000000)) == htonl(0xe0000000))



/* bitmaps */
#include "bitmaps/arrow_down"
#include "bitmaps/arrow_up"
#include "bitmaps/isc_icon"
#include "bitmaps/isc_mask"

/* media_[active=A|a] [rx=0|1] */
#include "bitmaps/audio_0A0"
#include "bitmaps/audio_1A0"
#include "bitmaps/audio_0A1"
#include "bitmaps/audio_1A1"
#include "bitmaps/audio_1r0"
#include "bitmaps/audio_0r0"
#include "bitmaps/audio_1r1"
#include "bitmaps/audio_0r1"
#include "bitmaps/video_1A0"
#include "bitmaps/video_0A0"
#include "bitmaps/video_1A1"
#include "bitmaps/video_0A1"
#include "bitmaps/video_1r0"
#include "bitmaps/video_0r0"
#include "bitmaps/video_1r1"
#include "bitmaps/video_0r1"
 
#include "bitmaps/picture_1A0"
#include "bitmaps/picture_0A0"
#include "bitmaps/picture_1A1"
#include "bitmaps/picture_0A1"
#include "bitmaps/picture_1r0"
#include "bitmaps/picture_0r0"
#include "bitmaps/picture_1r1"
#include "bitmaps/picture_0r1"
 
#include "bitmaps/netghost"

#include "bitmaps/audio"
#include "bitmaps/video"
#include "bitmaps/invite"
#include "bitmaps/add_media"
#include "bitmaps/help"
#include "bitmaps/configure"
#include "bitmaps/sip_handler"
#include "bitmaps/control"
#include "bitmaps/picture"
#include "bitmaps/reservation"
#include "bitmaps/ifloor"
#include "bitmaps/exit"
#include "bitmaps/vote"


/* audio control bitmaps */
#include "bitmaps/digital_in"
#include "bitmaps/digital_out"
#include "bitmaps/headphone_out"
#include "bitmaps/line_in"
#include "bitmaps/line_out"
#include "bitmaps/microphone_in"
#include "bitmaps/speaker_out"
#include "bitmaps/invalid_in"
#include "bitmaps/invalid_out"

#include "tcl/editor.c"
#include "tcl/mailcap.c"
#include "tcl/audioctl.c"
#include "tcl/application.c"
#include "tcl/blist.c"
#include "tcl/button_inc.c"
#include "tcl/clist.c"
#include "tcl/conference.c"
#include "tcl/config.c"
#include "tcl/mint.c"
#include "tcl/control_isc.c"
#include "tcl/control_sio.c"
#include "tcl/control_tcp.c"
#include "tcl/control_udp.c"
#include "tcl/field.c"
#include "tcl/fileselect_f.c"
#include "tcl/getopt.c"
#include "tcl/help.c"
#include "tcl/http.c"
#include "tcl/isc.c"
#include "tcl/invite.c"
#include "tcl/hscale.c"
#include "tcl/lremove.c"
#include "tcl/member.c"
#include "tcl/preference.c"
#include "tcl/isc_rsvp.c"
#include "tcl/play.c"
#include "tcl/record.c"
#include "tcl/session.c"
#include "tcl/rtp_avp.c"
#include "tcl/sip.c"
#include "tcl/sip_handler.c"
#include "tcl/sip_file.c"
#include "tcl/sip_media_list.c"
#include "tcl/sip_phonebook.c"
#include "tcl/sip_request.c"
#include "tcl/sip_response.c"
#include "tcl/statistics.c"
#include "tcl/img_trace.c"
#include "tcl/graph.c"
#include "tcl/tkerror_n.c"
#include "tcl/warning.c"
#include "tcl/exit.c"


#include "tcl/fccntrl.c"

#define TE(x) if (Tcl_Eval(interp, x) != TCL_OK) fprintf(stderr," failed %s",interp->result);;
#define TKB(x) Tk_DefineBitmap(interp, Tk_GetUid(#x), x##_bits, x##_width, x##_height)

char ret[100];
char* video_8to24(char* win,Tcl_Interp *interp)
{
  int depth,i;
  Colormap colormap;
  Window child_window;
  XSetWindowAttributes attributes;
  unsigned long valuemask = 0;
  Tk_Window mainWindow=Tk_MainWindow(interp);

  Visual *theVisual=Tk_GetVisual(interp,mainWindow,"best",&depth,&colormap);
  attributes.background_pixel = 0x0000000;
  attributes.border_pixel = 0x99999999;
  attributes.colormap = colormap;
  valuemask = CWBackPixel | CWBorderPixel | CWColormap ;
  child_window= XCreateWindow(Tk_Display(mainWindow),
			      strtoul(win,NULL,16),
			      0,0, 100, 100, 0,
			      depth,
			      InputOutput, 
			      theVisual, 
			      valuemask,&attributes); 
  XSelectInput(Tk_Display(mainWindow),child_window ,
	       ExposureMask | StructureNotifyMask);
  XMapWindow(Tk_Display(mainWindow),child_window);
  for(i=0;i<100;i++)
    ret[i]=0;
  sprintf(ret,"0x%x",(unsigned int)child_window);
  return ret;
}

int Tcl_video_8to24(ClientData client, Tcl_Interp *interp, int argc,
  char *argv[])
{
  if (argc != 2) {
    sprintf(interp->result, "%s: illegal # of args (%d)", argv[0], argc);
    return TCL_ERROR;
  }
 
  Tcl_SetResult(interp, video_8to24(argv[1],interp), TCL_STATIC);
  return TCL_OK;
} /* Tcl_video_8to24 */
 

char* host2dottednotation(char *host)
{
  struct in_addr in;
  register struct hostent *hep;

  /* Check whether this is a dotted decimal. */
  if (!host) {
    in.s_addr = INADDR_ANY;
  }
  else if ((in.s_addr = inet_addr(host)) != -1) {
  }
  /* Attempt to resolve host name via DNS. */
  else if ((hep = gethostbyname(host))) {
    in = *(struct in_addr *)(hep->h_addr_list[0]);
  }
  return inet_ntoa(in);
} /* host2ip */
        


int Tcl_host2dottednotation(ClientData client, Tcl_Interp *interp, int argc, char *argv[]
)
{
  char* result;

  if (argc != 2) {
    sprintf(interp->result, "%s: illegal # of args (%d)", argv[0], argc);
    return TCL_ERROR;
  }

  result = host2dottednotation(argv[1]);
  if (result > 0) {
    sprintf(interp->result, "%s", result);
  } else {
    sprintf(interp->result, "error");
  }

  return TCL_OK;
} /* Tcl_ */
int check_multicast(char *hostaddr)
{

  struct hostent* host_ent;
  struct in_addr inaddr;

  host_ent = gethostbyname(hostaddr);

  memcpy(&(inaddr.s_addr),host_ent->h_addr_list[0], sizeof(long));

  return (int)MULTICAST(inaddr.s_addr);

} /* host2ip */
        


int Tcl_conn_is_multicast(ClientData client, Tcl_Interp *interp, int argc, char *argv[]
)
{

  int i=0;

  i = check_multicast(argv[1]);  

  sprintf(interp->result, "%d", i);

  return TCL_OK;
} /* Tcl_conn_is_multicast*/


int Isc_Init(Tcl_Interp *interp)
{
  static struct {
    char *name;
    char *var;
    char *def;
  } libraries[] = {
    {"ISC_LIBRARY",    "isc_library",   "."},
    {0,0}
  };
 
  extern int Tcl_idstring(ClientData clientData, Tcl_Interp *interp,
			  int argc, char **argv);
  extern int Tcl_host2ip(ClientData clientData, Tcl_Interp *interp,
			  int argc, char **argv);
  extern int noperiod_cmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char **argv);
  extern int nospace_cmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char **argv);
  extern int Tcl_getclock(ClientData clientData, Tcl_Interp *interp,
			  int argc, char **argv);
  extern int Tcl_strftime(ClientData clientData, Tcl_Interp *interp,
			  int argc, char **argv);
  extern int source_file(Tcl_Interp *interp, char *file);
  extern int Tcl_SetSockBuffer_cmd(ClientData clientData, Tcl_Interp *interp, 
			  int argc, char **argv);

  extern char *prog_name;

  int i;
  Tcl_SetVar(interp, "prog_name",prog_name, TCL_GLOBAL_ONLY);


  /*
   * Call Tcl_CreateCommand for application-specific commands, if
   * they weren't already created by the init procedures called above.
   */
  Tcl_CreateCommand(interp, "video_8to24", Tcl_video_8to24, 0, 0);
  Tcl_CreateCommand(interp, "getdottedaddr", Tcl_host2dottednotation, 0, 0);
  Tcl_CreateCommand(interp, "connection_is_multicast", Tcl_conn_is_multicast, 0, 0);
  Tcl_CreateCommand(interp, "host2ip", Tcl_host2ip, 0, 0);
  Tcl_CreateCommand(interp, "idstring", Tcl_idstring, 0, 0);
  Tcl_CreateCommand(interp, "noperiod", noperiod_cmd, 0, 0);
  Tcl_CreateCommand(interp, "nospace",  nospace_cmd,  0, 0); 
  Tcl_CreateCommand(interp, "getclock", Tcl_getclock, 0, 0); 
  Tcl_CreateCommand(interp, "strftime", Tcl_strftime, 0, 0); 
  Tcl_CreateCommand(interp, "Tcl_SetSockBuffer", Tcl_SetSockBuffer_cmd, 0, 0);

  /* set up bitmaps */
  Tk_DefineBitmap(interp, Tk_GetUid("arrow_down"), arrow_down_bits,
		  arrow_down_width, arrow_down_height);
  Tk_DefineBitmap(interp, Tk_GetUid("arrow_up"), arrow_up_bits,
		  arrow_up_width, arrow_up_height);
  Tk_DefineBitmap(interp, Tk_GetUid("isc_icon"), isc_icon_bits,
		  isc_icon_width, isc_icon_height);
  Tk_DefineBitmap(interp, Tk_GetUid("isc_mask"), isc_mask_bits,
		  isc_mask_width, isc_mask_height);

  /* member indication */
  TKB(audio_1A0);
  TKB(audio_1A1);
  TKB(audio_1r0);
  TKB(audio_1r1);
  TKB(video_1A0);
  TKB(video_1A1);
  TKB(video_1r0);
  TKB(video_1r1);
  TKB(picture_1A0);
  TKB(picture_1A1);
  TKB(picture_1r0);
  TKB(picture_1r1);
  TKB(netghost);

  TKB(audio_0A0);
  TKB(audio_0A1);
  TKB(audio_0r0);
  TKB(audio_0r1);
  TKB(video_0A0);
  TKB(video_0A1);
  TKB(video_0r0);
  TKB(video_0r1);
  TKB(picture_0A0);
  TKB(picture_0A1);
  TKB(picture_0r0);
  TKB(picture_0r1);


  TKB(audio);
  TKB(video);
  TKB(invite);
  TKB(add_media);
  TKB(help);
  TKB(configure);
  TKB(sip_handler);
  TKB(control);
  TKB(picture);
  TKB(reservation);
  TKB(ifloor);
  TKB(exit);
  TKB(vote);

  TKB(digital_in);
  TKB(digital_out);
  TKB(headphone_out);
  TKB(line_in);
  TKB(line_out);
  TKB(microphone_in);
  TKB(speaker_out);
  TKB(invalid_in);
  TKB(invalid_out);


  /* set up library directories */
  for (i = 0; libraries[i].name; i++) {
    char *lib_dir = getenv(libraries[i].name);
    if (lib_dir == NULL) lib_dir = libraries[i].def;
    Tcl_SetVar(interp, libraries[i].var, lib_dir, TCL_GLOBAL_ONLY);
  }
  
  if (source_file(interp, "$isc_library/isc.tcl") < 0) {
    TE(editor_tcl)
    TE(mailcap_tcl)
    TE(application_tcl)
    TE(audioctl_tcl)
    TE(mint_tcl)
    TE(blist_tcl)
    TE(button_inc_tcl)
    TE(clist_tcl)
    TE(conference_tcl)
    TE(config_tcl)
    TE(control_isc_tcl)
    TE(control_sio_tcl)
    TE(control_tcp_tcl)
    TE(control_udp_tcl)
    TE(field_tcl)
    TE(fileselect_f_tcl)
    TE(getopt_tcl)
    TE(graph_tcl)
    TE(help_tcl)
    TE(hscale_tcl)
    TE(http_tcl)
    TE(img_trace_tcl)
    TE(invite_tcl)
    TE(lremove_tcl)
    TE(isc_rsvp_tcl)
    TE(member_tcl)
    TE(preference_tcl)
    TE(play_tcl)
    TE(record_tcl)
    TE(rtp_avp_tcl) 
    TE(session_tcl)
    TE(sip_tcl) 
    TE(sip_handler_tcl) 
    TE(sip_file_tcl) 
    TE(sip_media_list_tcl) 
    TE(sip_phonebook_tcl) 
    TE(sip_request_tcl) 
    TE(sip_response_tcl)  
    TE(statistics_tcl)
    TE(tkerror_n_tcl)
    TE(warning_tcl) 
    TE(exit_tcl)
    TE(isc_tcl)
  }
   
  return TCL_OK;    
}
