#
# rtp_avp.tcl
# procedures to convert media encoding to the appropriate RTP profile 
# and vice-versa
#



#
# create a list to match media encodings to RTP profiles
# fromat: {"encoding [channels clock-rate]" RTP_profile}
#
proc rtp_avp_load {} {
  global rtp_avp_list
 
  set rtp_avp_list { \
    {"pcmu 1 8000" 0} \
    {"gsm 1 8000" 3} \
    {"dvi4 1 8000" 5} \
    {"lpc 1 8000" 7} \
    {"dvi16 1 16000" 6} \
    {"l16_2 2 44100" 10} \
    {"l16_1 1 44100" 11} \
    {cellb 25} \
    {jpeg 26} \
    {h261 31} \
    {mpeg 32} \
  }
} ;# rtp_avp_load



#
# convert RTP profile to media encoding
# - in: profile
# - return : {encoding [channels clock_rate]}
#
proc get_code {profile} {
  global rtp_avp_list

  # extracts encoding out of rtp profile
  rtp_avp_load
  foreach line $rtp_avp_list {
    if {[lindex $line 1] == $profile} {
      return [lindex $line 0]
    }
  }
  return
} ;# get_code



#
# convert media encoding to RTP profile
# - in: encoding
# - return : RTP profile
#
proc get_profile {encoding} {
  global rtp_avp_list

  # extracts rtp/vat profile out of encoding
  rtp_avp_load
  foreach line $rtp_avp_list {
    if {[lindex [lindex $line 0] 0] == $encoding} {
      return [lindex $line 1]
    }
  }
  return
} ;# get_profile

