# This file contains a default version of the tkError procedure.  It
# posts a dialog box with the error message and gives the user a chance
# to see a more detailed stack trace.
#
# $Id: tkerror_n.tcl,v 1.2 1998/07/20 16:41:43 stb Exp $
#

proc tkerror err {
  global tk_library

  if [info exists tk_library] {tkerror.nevot $err} else {error $info}
}

proc tkerror.nevot err {
  global errorInfo errorCode errorFatal prog_name

  set info $errorInfo

  # if no colors left, display message only (Tk will dump core
  # otherwise)
  if [string match "no more colors*" $err] {
    puts stderr "Error: $err"
    return
  }

  # if window exists, don't try to put up another one
  if {[winfo exists .tkerrorDialog]} {
    puts stderr "Error: $err"
    return
  }

  if {[info exists errorFatal] && $errorFatal} {
    set b [tk_dialog .tkerrorDialog "Error in Tcl Script $prog_name ([pid])" \
      "$err" error 0 "Abort" "See Stack Trace"]
    if {$b == 1} {set b 2}
  } else {
    set b [tk_dialog .tkerrorDialog "Error in Tcl Script $prog_name ([pid])" \
	       "$err" warning 1 "Abort" "Continue" "See Stack Trace"]
  }

  if {$b == 0} exit

  if {$b == 1} return

  if {$b == 2} {
    # if OK, display stack trace
    set w .tkerrorTrace
    catch {destroy $w}
    toplevel $w -class ErrorTrace
    wm minsize $w 1 1
    wm title $w "Stack Trace for Error"
    wm iconname $w "Stack Trace"
    button $w.ok -text OK -command "tkerror_ok $w $b"
    text $w.text -relief raised -bd 2 -yscrollcommand "$w.scroll set" \
  	    -setgrid true -width 40 -height 10
    scrollbar $w.scroll -relief flat -command "$w.text yview"
    pack $w.ok -side bottom -padx 3m -pady 3m -ipadx 2m -ipady 1m
    pack $w.scroll -side right -fill y
    pack $w.text -side left -expand yes -fill both
    $w.text insert 0.0 $info
    $w.text mark set insert 0.0

    # Center the window on the screen.

    wm withdraw $w
    update idletasks
    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
	    - [winfo vrootx [winfo parent $w]]]
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
	    - [winfo vrooty [winfo parent $w]]]
    wm geom $w +$x+$y
    wm deiconify $w
  }
}
proc tkwarning {err {param 1}} {
  global errorInfo errorCode errorFatal prog_name
  set info $errorInfo

  # if no colors left, display message only (Tk will dump core
  # otherwise)
  if [string match "no more colors*" $err] {
    puts stderr "Error: $err"
    return 0
  }

  # if window exists, don't try to put up another one
  if {[winfo exists .tkerrorDialog]} {
    puts stderr "Error: $err"
    return 0
  }

  if {$param == 1} {
    set b [tk_dialog .tkerrorDialog "Warning in  $prog_name ([pid])" \
      "$err" warning 0 "Continue"]
    return 1
  }
  if {$param == 2} {
    set b [tk_dialog .tkerrorDialog "Warning in  $prog_name ([pid])" \
      "$err" warning 0  "Ignore" "Accept"]
      if { $b==0 } { return 1 }
      if { $b==1 } { return 2 }
  }
}

proc tkerror_ok {w b} {
  global errorFatal

  if {[info exists errorFatal] && $errorFatal} exit
  destroy $w
}
