/* Filename: ac.h         
 *   Author: Dorgham Sisalem  
 *  Purpose: Global Definitions               */

/*
 * Copyright (c) 1996 Dorgham Sisalem
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice appears in all copies of this software.
 *  See the copyright file for copyright information*
 */

#ifndef _ac_h
#define _ac_h

#define WORST_RECEIVER 0
#define CONGESTION_RATIO 1
#define LOSS_DELAY_BASED_ADJUSTMENT 3
#define LOSS_BASED_ADJUSTMENT 2
#define DIRECT_ADJUSTMENT 5

#define CONTROL                    10
#define ALGORITHM                  11
#define ALPHA                      12
#define MAX_BANDWIDTH              13
#define MIN_BANDWIDTH              14
#define MULTIPLICATIVE_DECREASE    15
#define ADDITATIVE_DECREASE        16
#define LOSS_LOWER_THRESHOLD       17
#define LOSS_UPPER_THRESHOLD       18
#define CONGESTED_RECV_FRACTION    19
#define LOADED_RECV_FRACTION       20

#define UPDATE_INTERVAL  5000000.0  /* the size of an update interval in usec */

typedef enum {UNLOADED, LOADED, CONGESTED} network_state_t;



typedef struct {
  double filtered_loss;     /* linear filtered loss */
  double loss;              /* last reported loss  0<= loss <=1 */
  int ssrc;                 /* actual receiver ssrc */

  double current_Q;         /* last measured queuing delay */
  double max_Q;             /* last seen queueing delay before a loss */
  double rtt;               /* last measured round trip time */
  double min_rtt;           /* smalles measured round trip time, is 
			       considered as the propagation delay */
  u_int32 time;                 /* update time of this member in usecs */
  double TCP_Rate;          /* TCP rate for loss and min_rtt */
  double bbw;               /* bottleneck bandwdith for thos member */
} mem_t;


/*
 * session structure: contains all information about the actual session
*/
typedef struct {
  u_int32 ssrc;                      /* ssrc of controlled sender */
  int (*analyse)();
  int (*adjust)();
  Tcl_HashTable member_table;        /* hash table for members */
  int members;                       /* # of members including the sender */

  /* adjustment specific variables */
  int control;
  int min_bandwidth;                /* minimal bw value under which no control
				     * should be used */
  int max_bandwidth;                /* maximal bw value under which no control
				     * should be used */
  int allowed_bandwidth;             /* calculated bw to use */
  double alpha;                      /* used for the filter */
  double multiplicative_decrease;    /* value with which to decrease the bw */
  double additive_increase;          /* initial value with which to increase 
					bw */
  double aif;                        /* current additive_increase factor*/
  double loss_lower_threshold;      
  double loss_upper_threshold;
  double congested_recv_fraction;
  double loaded_recv_fraction;
  int algorithm;                    /* WORST_RECEIVER or CONGESTION_RATIO */
  double loss;                      /* the heighest loss seen in RR*/
  u_int32 ssrc_loss;                /* ssrc of the source reporting the 
				     * heighets loss */
  int time;                         /* time of last update interval in usec*/
} ac_t;

/*
 * An SSRC database entry.  Holds all the information about a given SSRC.
 * contains, a list of payload-type -> local decoder translations, in order
 * to facilitate parsing of RTP data streams from this SSRC.  The application
 * program returns (with the fmt_decoder routine) an opaque identifier which
 * will be used later to identify the packet type that this SSRC is sending.
 */



/* external function declarations */



/* initialize the needed information structure for the application 
 * controller 
 */
extern void ac_init(u_int32 ssrc);

/* calculate the appropriate bandwidth to use */
extern int ac_qos_control(int ssrc, double loss, double rtt,int bandwidth);

/* min_bandwidth is set by default to 0. In case of using some kind of 
 * guranteed reservation the min_bandwidth value can be accordungly changed
 */
extern void ac_update(int tag, int parameter);
extern int ac_max_bandwidth(); /* return the max bandwidth value of the controller */

/* delete the entry made for the stream with the ssrc value */
extern void ac_delete_entry(u_int32 ssrc);

/* update the bbw vale */
extern void ac_update_bbw(int ssrc,double bbw);

#endif

