#ifndef color_map_h
#define color_map_h

#define GREY_LEVELS 32
#define YUVLIM(x) (((x) < 0) ? 0 : ((x) > 127) ? 127 : (x))


/*
 * Classes for mapping YUV image data into a representation that
 * can be rendered on the display.  The ColorModel class represents
 * a finite-element color cube, which might be uniformaly sampled,
 * or biased; or, a map from YUV space to 24-bit true color.
 * The Colorizer class represents the agents that do the work of
 * mapping from YUV space to the ColorModel space.  For example,
 * they might do dithering or simple (non-uniform) quantization.
 */

	typedef struct{
		short r;
		short g;
		short b;
		short y;
		short u;
		short v;
	}color;
	color colors_[256];
	/*inline const color& colors(int n) const { return (colors_[pixel_[n]]); }*/

	
	/*void setvisual(Visual* v, Colormap c, int d) {
		visual_ = v;
		colormap_ = c;
		depth_ = d;
	}*/
	Display* dpy_;
	Colormap colormap_;
	Visual* visual_;
        Window win_;
	int depth_;

	int ncolor_;
	u_long pixel_[256];
	u_short grey_lut_[512];
        const char* file_;
        unsigned long* lut_;

extern int allocate_colormap(Display *theDisp, Visual* theVisual, Window win,Colormap theCmap);

extern void free_color();
extern void cm_installcolor(); /* use the private color map */
extern void cm_uninstallcolor(); /* revert to the standard color map */

 

#endif
