/* 
 * vic_jpeg.h
 *
 * Copyright (c) 1996 by Dorgham Sisalem GMD fokus all rights reserved.
 *
 * The code presented here is ported into C by the author from
 * the code developed by the University of California in C++. The code
 * is part of the vic video tool developed by the University.
 *
 * Copyright (c) 1993-1995 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Dorgham Sisalem
 *       GMD Fokus
 * 4. Neither the name of the GMD Fokus nor of the author may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <X11/Xlib.h>
/*XXX*/
#define NUM_HUFF_TBLS       4	/* Huffman tables are numbered 0..3 */
#define BITS_NUM 8


	/*
	 * Baseline JPEG decompression parameters
	 * (which is the only mode we can handle).
	 * (describes the JPEG input stream)
	 */
typedef struct  {
		int hsf;	/* horizontal sampling factor (1..4) */
		int vsf;	/* vertical sampling factor (1..4) */
		int qno;	/* quantization table selector (0..3) */
		/* for decompression, they are read from the SOS marker. */
		int dc_tbl_no;	/* DC entropy table 2selector (0..3) */
		int ac_tbl_no;	/* AC entropy table selector (0..3) */
		int dc;		/* dc predictor */
	}component;
typedef struct{
		int width;
		int height;
		int precision;
		struct {
			int id;
			int hsf; /* horiz sample factor */
			int vsf; /* vertical sample factor */
			int qno;
			/*
			 * XXX Assume all scans the same.
			 */
			int ac_tbl_no;
			int dc_tbl_no;
		} comp[4];
		int qtab[4][64];
		  u_char *ac_huffbits[4];
		  u_char *ac_huffval[4];
		  u_char *dc_huffbits[4];
		  u_char *dc_huffval[4];
	  u_char *inb_;		/* input buffer XXX make u_long */
	  u_char *end_;
	u_int bb_;		/* bit buffer */
	int nbb_;		/* # bits in bit buffer */

	int size_;
	int ncomp;
	component comp_[3];
        int quality;
        int framesize_;
	int rlen_;		/* restart interval length (0 if not using) */
	int rcnt_;		/* restart interval counter */
	int nblk_;		/* number of 8x8 blocks*/
	u_char* rvts_;
	int ren_now_,dec_now_;
        u_int uvtab_[65536],omask_,pmask_;


	/*
	 * Quantization tables.  qt_ is always stored in 
	 * natural order (i.e., not zigzagged).  fqt_ is
	 * transformed by the DCT module to be in whatever
	 * form it finds most convenient.
	 */
	int qt_[4][64];
	int fqt_[4][64];

	/*
	 * Huffman tables.
	 */
	u_short* dcht_[NUM_HUFF_TBLS];
	u_short* acht_[NUM_HUFF_TBLS];
	int ncol_;		/* no. of mcu's across scan */
	int nrow_;		/* no. of mcu's down scan */


	int color_;

	/* state for suppressing work when blocks don't change much */
	int thresh_;
	int cthresh_;

	int rcrop_;
	int lcrop_;
	int topcrop_;
	int botcrop_;

        char (*dither) (char *, u_int, u_int, u_int, u_int, char*); 
                     /* which dithering algorithm to use: 24 or 8 bits */
        char (*render) (char* ,char*,char*);
        char (*dither_down) (char *, u_int, u_int, u_int, u_int, char*); 
                     /* which dithering algorithm to use: 24 or 8 bits */
        char (*render_down) (char* ,char*,char*);
	u_char* frm_;
	short* cache_;
	}config;
typedef struct{
		int top;
		int right;
		int left;
		int ytopskip;
		int uvtopskip;
		int marktopskip;
		int ylskip;
		int uvlskip;
		int marklskip;
		int yrskip;
		int uvrskip;
		int markrskip;

		int ftopskip;
		int flskip;
		int frskip;

	} margin;

/****************************vic_jpeg.c******************************/
extern void init_JpegDecoder_411(config* c);
extern int quality_to_qfactor(int v);
extern void jpeg_qt(int q, int* out,   int* in);
extern void defaults(config* c);
extern void quantizer(config* c);
extern int q_to_thresh(int q);
extern void init(config* c);
extern int compute_margins(int maxh, int maxv,config *c);
extern void setlrskips(config* c);
extern void freehufftab(config* c);
extern int zag_blkdiff(short* blk, short* ref);
extern int inorder_blkdiff(short* blk, short* ref);
extern int blkdiff(short* blk, short* cache);
extern int rdqt(  u_char* p,config *c);
extern void restart(config* conf);
extern u_char * parseJFIF(  u_char* in,config *c);
extern void Rdct(int nc, register short *bp, u_int m0, u_int m1,
	  u_char* pixels, int stride, int qno);
extern int huffparse(component* p, short* out, short* ref,
	      u_int* mask, int dontskip);
extern int huffskip(component* p);
extern void huffreset(config* c);
extern u_short* huffbuild(  u_char* bits,   u_char* vals);

/****************** EXPORTED JPEG PROCEDURES ************************/
extern char * Init_JpegDecoder(int width, int height,int quality,char* display,
			       int window,char* the_Visual);
extern char * Delete_JpegDecoder(char* confi);
extern int Jpeg_AllocColors(char* c,char* the_Visual,char* the_Disp,int scale);
extern void JpegDecoder_411(u_char* in,int len,int q,char* c,char* the_Image);
extern void Jpeg_ReInit(char *c, char* theDisp, int window);
extern int alloc_true_colors(char* c,char* the_Visual, char* theDisp);
extern void  map_411(config* c,u_int off,u_int x,u_int width,
		     u_int height,char* the_Image);
extern void dither_411(config* c,
			u_int off, u_int x,
			u_int width, u_int height,char* the_Image);
extern void renderer(char *buf,char *c,char* the_Image);
extern void init_jpeg_q(char *c, int q);













