#include "copyright.h"

/*
 * contains all the necessary procedures for creating displaying and deleting
 * video widgets
 *
 * We maintain a video map of maximally WINDOW_WIDTH x WINDOW_LENGTH window 
 * slots. For each new data stream a new slot is allocated and the video
 * is enalrged. If a steam is no longer being displayed, either because the
 * sender left the session or the receiver does not want to see this stream
 * the video slot is cleared and declared as a "no entry" slot. However, the 
 * allocated resources for this slot -image, GC, shared memory segment- are
 * not freed but kept for the case a new stream joins the session. This would 
 * then reduce the time needed for initializing the display of new streams
 *
 * the first 6 sterams can use shared memory segments to speed up the display
 *
 *  See the copyright file for copyright information
 */

static char rcsid[]  = "$Id: xv.c,v 1.6 1998/09/01 15:37:42 dor Exp $";


#include <tk.h>
#include "xv.h"
#include "color_map.h" 
#include <sys/utsname.h> /* uname() */
#include <X11/Xatom.h>
#include "types.h"
#include "rtp.h"
#include "nevit.h"
#include "x11_grab.h"


/* set the requirements for the shared memory segments */
#include <sys/shm.h>
#if defined(sun) && !defined(__svr4__)
/*int shmget(key_t, int, int);*/
int shmdt(char*);
int shmctl(int, int, struct shmid_ds*);
#endif
#ifdef __osf__
int XShmGetEventBase(struct _XDisplay *);
#else
int XShmGetEventBase(Display *);
#endif
#ifdef sgi
#define XShmAttach __XShmAttach__
#define XShmDetach __XShmDetach__
#define XShmPutImage __XShmPutImage__
#endif
#include <X11/extensions/XShm.h>
#ifdef sgi
#undef XShmAttach
#undef XShmDetach
#undef XShmPutImage
int XShmAttach(Display*, XShmSegmentInfo*);
int XShmDetach(Display*, XShmSegmentInfo*);
int XShmPutImage(Display*, Drawable, GC, XImage*, int, int, int, int,
                 int, int, int);
#endif

XShmSegmentInfo *shminfo_;
static struct utsname u_name;     /* uname value */
int USE_SHM=0;

#undef SIGCHLD           /* defined in both Xos.h and signal.h */
#include <signal.h>      /* for interrupt handling */


#define WINDOW_WIDTH 4  /*the horizontal number of video images*/
#define WINDOW_LENGTH 3  /* the vertical number of video images */      
#define MAX_SHM 6        /* maximum number od shared memory segments that 
			  * can be used */


/* decalre the needed varialbles */

Tk_Window              slave[WINDOW_LENGTH];
XImage                 *theImage;
GC                     theGC;
Window                 mainW;
static Visual          *theVisual;
static Tk_Window       containerWindow=0;
static int first=0;

extern Tcl_Interp      *interp_;
Tk_Window              mainWindow;

video_map_struct       video_map[WINDOW_LENGTH*WINDOW_WIDTH];

   
/************* Local procedures declaration ************************/
void clear_video_map(int window);
extern void free_colors();

/**************************************************************************/
/*
 * Redraw a window slot when there are no Tk events to be processed 
 */
void draw_when_idle(ClientData place)
{
  int data=(int) place;
  drawWindow(0,0, video_map[data].theImage->width*s->scale_factor,
	     video_map[data].theImage->height*s->scale_factor,
	     (char *)video_map[data].theImage, (char *)video_map[data].theDisp,
	     (char *)video_map[data].theGC,
	     (char *)Tk_WindowId(video_map[data].childWindow),
	     video_map[data].name,(char *)video_map[data].shminfo);
}

/**************************************************************************/
/*
 * clear an empty slot when there are no Tk events to be processed 
 * and note it as a "no entry" slot
 */
void clear_when_idle(ClientData place)
{
  int data=(int) place;
  XFillRectangle(video_map[data].theDisp,
		 Tk_WindowId(video_map[data].childWindow),
		 video_map[data].theGC_o,0,0,
		 video_map[data].theImage->width*s->scale_factor,
		 (video_map[data].theImage->height*s->scale_factor)+40);
  XDrawString(video_map[data].theDisp,Tk_WindowId(video_map[data].childWindow),
	      video_map[data].theGC,(video_map[data].theImage->width/2)-16,
	      video_map[data].theImage->height+20,
	      "   no  entry    ",16);  
}

  
/**************************************************************************/
/*
 * handle the Exposure and Destroy events of the single video slots 
 */
void Event_Proc(ClientData place, XEvent *eventPtr)
{
  int data=(int) place;
  switch (eventPtr->type) {
  case Expose:
    if(data>0) { /* only for child windows */
      if(video_map[data].busy) {
	Tcl_DoWhenIdle((Tcl_IdleProc *)draw_when_idle,(ClientData) data);
      }
      else 
	if(s->scale_factor==1)
	  Tcl_DoWhenIdle((Tcl_IdleProc *)clear_when_idle,(ClientData) data);
    }
    break;
  case DestroyNotify:
    raise(SIGINT);
    break;
  case EnterNotify:
    cm_installcolor();
    break;
  case LeaveNotify:
    cm_uninstallcolor();
    break;  
  case ResizeRequest:
    break;
  }  
}

/*******************************************************************/
/* handle some window events -out of some strange reason the leave 
 * event was not catched by X so we use TK
 */
void events_Cmd(ClientData clientData, Tcl_Interp *interp,
                           int argc, char *argv[])
{
  switch (atoi(argv[1])) {
  case 1:  /* Leave */
    cm_uninstallcolor();
    break;  
  }    
}	  

/**************************************************************************/
/* resize the window after clicking on the pictures. either up or down */
int resize_Cmd(ClientData clientData, Tcl_Interp *interp,
                           int argc, char *argv[])
{
  static int resizeing=1;
  clear_TkAll();
  if(resizeing){
    resizeing=0;
    s->scale_factor=2;
    restart_Tk_up(atoi(argv[1]),s->scale_factor);/*set a restart flag so active
						   stream starts
						   building a new slot*/
  }
  else{
    resizeing=1;
    s->scale_factor=1;
    restart_Tk_down(s->scale_factor);
  }
  return TCL_OK;
}


/**************************************************************************/
void restart_Tk_up(int window,int scale)
{
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  member_t *m=NULL;
  
  clear_video_map(0);
  /* find the member to diplay and create a window for him*/
  for (entryPtr = Tcl_FirstHashEntry(&s->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (member_t *) Tcl_GetHashValue(entryPtr);
    if(m->cname!=NULL){
      if((m->cname!=NULL)&&!strcmp(m->cname,video_map[window].name)){
	if(m->preview>0)
	  m->preview=1;
	do_TkStuff((int *)&(s->active.theImage),(int *)&(s->active.theDisp),
		   (int *)&(s->active.theGC),(int *)&(s->active.mainW),
		   (int *)&(s->active.theVisual),
		   m->width,m->height,
		   (int *)&(s->active.shminfo), m->cname, m->xil_stuff);
      break;
      }
    }
  }
  /* set all active window to inactive ->preview =2 */
  for (entryPtr = Tcl_FirstHashEntry(&s->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (member_t *) Tcl_GetHashValue(entryPtr);
    if(m->cname!=NULL){
      if(!strcmp(m->cname,video_map[window].name)){
	if(m->preview>0)
	  m->preview=1;
      }
      else{
	if(m->preview>0){
	  m->preview   = scale;
	}
      }
      /* ensure that the streams requireing xil stuff always have a window */
      m->mainW     = s->active.mainW;
      if(m->preview>0){
	m->theImage  = s->active.theImage;
	m->theDisp   = s->active.theDisp;
	m->theGC     = s->active.theGC;
	m->mainW     = s->active.mainW;
	m->theVisual = s->active.theVisual;
	m->shminfo   = s->active.shminfo;
	/* initialize the dither and render procedures */
	if(m->decoder!=NULL){
	  m->redecode((char *)m->decoder,(char *)m->theDisp,(int)m->mainW);
	  m->alloc_colors(m->decoder,(char *)m->theVisual,
		       (char *)m->theDisp,scale);
	}
      }
    }
  }

}
  
/***************************************************************************/
/* recover the displayed streams just as was the case before the resizing*/
void restart_Tk_down(int scale)
{
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  member_t *m=NULL;

  clear_video_map(0);
  for (entryPtr = Tcl_FirstHashEntry(&s->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (member_t *) Tcl_GetHashValue(entryPtr);
    /* there are different cases to be considered here:
       1. streams that were previewed before the resize and are previewd now
       2. streams that were not previewd before but should now
       3. streams that were previewd before but not anymore
       */
    if((m->decoder!=NULL)&&m->preview>0) { /* case 1 */
      do_TkStuff((int *)&(m->theImage),(int *)&(m->theDisp),
		 (int *)&(m->theGC),(int *)&(m->mainW),
		 (int *)&(m->theVisual),
		 m->width,
		 m->height,
		 (int *)&(m->shminfo), m->cname, m->xil_stuff);
      m->redecode(m->decoder,(char *)m->theDisp,(int)m->mainW);
      m->preview=1;
      m->alloc_colors(m->decoder,(char *)m->theVisual,(char *)m->theDisp,
		      scale);
    }
    if(m->preview>0) /* set all members allowed to preview to 1 case 2*/
      m->preview=1;
    if(m->decoder!=NULL){/*reinitialize the render and dither functions case3*/
      m->alloc_colors(m->decoder,(char *)m->theVisual,(char *)m->theDisp,
		      scale);
    } 
  }
}



/***********************************************************************
 *
 * if the dimensions of a stream changed end the display and restart it 
 * from new. This was the easiest way to avoid the confusion of images
 * of different sizes
 *
 ************************************************************************/

void restart_TKStuff()
{
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  member_t *m=NULL;

  end_TkAll(1);
  for (entryPtr = Tcl_FirstHashEntry(&s->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (member_t *) Tcl_GetHashValue(entryPtr);
    /*  streams that were previewed before the change and are previewd now */
    if(m->preview==1) { /* display this stream */
      do_TkStuff((int *)&(m->theImage),(int *)&(m->theDisp),
		 (int *)&(m->theGC),(int *)&(m->mainW),
		 (int *)&(m->theVisual),
		 m->width,
		 m->height,
		 (int *)&(m->shminfo), m->cname, m->xil_stuff);
    }
    /* reinitialize all other streams */
    if(m->decoder != NULL) {
      free(m->decoder);
      m->decoder=NULL;
      m->decoder=m->init_decompression(m->width,m->height,50,
				       (char *)m->theDisp,
				       (int)m->mainW,(char *)m->theVisual);
    
      m->alloc_colors(m->decoder,(char *)m->theVisual,(char *)m->theDisp,
		      s->scale_factor);
    }
    
  }
}

/**************************************************************************/
/*
 * initialize the video map by setting all the entries of video slots to NULL
 * and creating a container window which is the parent of all video slots
 * return -1 error
 * return  1 OK
 */
int init_video_map()
{
  int i,l, depth;
  Colormap cm=1;

  if(containerWindow!=0)
    return ((int)containerWindow);
  for(i=0;i<WINDOW_LENGTH;i++)
    slave[i]=NULL;
 
  for(i=0;i<WINDOW_LENGTH*WINDOW_WIDTH;i++) {
    video_map[i].theImage      =   NULL;
    video_map[i].theGC         =   NULL;
    video_map[i].mainW         =   0;
    video_map[i].theVisual     =   NULL;
    video_map[i].theDisp       =   NULL;
    video_map[i].shminfo       =   NULL;
    video_map[i].name          =   NULL;
    video_map[i].restart       =   0   ;
    video_map[i].xilW          =   0   ;
    for(l=0;l<20;l++)
      video_map[i].child_name[l]     =  '0' ; 
    for(l=0;l<100;l++)
      video_map[i].command[l]        =  '0' ;
    for(l=0;l<20;l++)
      video_map[i].where[l]          =  '0' ;
    video_map[i].busy                =   0  ;
    video_map[i].width               =   0  ;
  }
  /* check if main window exists */
  mainWindow=Tk_MainWindow(interp_);
  Tk_MapWindow(mainWindow);
  if(!Tk_IsMapped(mainWindow))
    Tk_MakeWindowExist(mainWindow);
  theVisual=Tk_Visual(mainWindow);
  if(!containerWindow){
    if((containerWindow=  /* create the container window */
	Tk_CreateWindowFromPath(interp_, mainWindow,".container",NULL))==NULL){
      fprintf(stderr," %s \n",interp_->result);
      perror("could not open container window");
       return -1;
    }
  } 
  if((theVisual=Tk_GetVisual(interp_,mainWindow,"best",&depth,&cm))==NULL) {
    fprintf(stderr,"%s \n",interp_->result);
    return -1;
  }
  if(!Tk_SetWindowVisual(containerWindow, theVisual, depth,cm )) {
    perror(" could not change visual"); 
    return -1;
  }
  BACKGROUND=0x000;
  FORGROUND=0x999999;
  if(theVisual->class ==PseudoColor){
      BACKGROUND=555555;
      FORGROUND =00000;   
  }
  /* look for a TrueColor, 24-bit or more (pref 24) and save the appropriate
   * visual information
   */


  Tk_SetWindowBackground(containerWindow,FORGROUND);
  Tk_GeometryRequest(containerWindow, 150,0);
  Tk_SetClass(containerWindow, "NeViT");    
    
  /* map the container window on the screen */
  Tk_MapWindow(containerWindow);
  Tcl_Eval(interp_,"pack .container  -padx 1 -pady 1 -ipadx 1 -ipady 1 -expand yes -fill both");
  XSynchronize(Tk_Display(mainWindow),True);
  Tk_CreateEventHandler(mainWindow,
			LeaveNotify|ResizeRequest|DestroyNotify|EnterNotify|FocusIn|FocusOut|GraphicsExpose|VisibilityNotify|MapNotify|UnmapNotify, 
			Event_Proc, (ClientData) -1); 
  { /* bind some events to this window */
    char command[50];
    sprintf(command,"bind . <Leave> {tk_events 1}");
    Tcl_Eval(interp_,command);
  }
  /* handle the case of pseudo color environments by allocating 
   * a private color map for the main window and installing it*/
  if(theVisual->class ==PseudoColor){
    if(allocate_colormap(Tk_Display(containerWindow),theVisual,
			 Tk_WindowId(containerWindow),cm)<0) {
      Exit();
    }
  }
  /* create tcl command for resizing the window */
  Tcl_CreateCommand(interp_, "resize", resize_Cmd,0, NULL);
  Tcl_CreateCommand(interp_, "tk_events", events_Cmd,0, NULL);
  return((int)containerWindow);
}


/**************************************************************************/
/*
 * search the vdieo map to find an empty slot are create a new one 
 * return # entry of the map
 */
int get_map(char *name,int width)
{
  int i,count=0;
  char command[50];

  /* search the video map for empty entries */
  for(i=0;i<=WINDOW_LENGTH*WINDOW_WIDTH;i++) {
    if(video_map[i].name!=NULL)
    /* case 1: found a slot that this stream has been already using
     * case 2: found an empy slot with the same dimensions as the stream
     * case 3: found an empty slot but with different dimensions so clear it 
     *         reuse it
     */
    if((video_map[i].name!=NULL)&&!strcmp(name,video_map[i].name)) {/* case 1*/
      video_map[i].width=width;
      video_map[i].busy=1;  /* found a slot that this stream has already been
			     * using */
      break;
    }
    /* case 2*/
    if(!video_map[i].busy&&((video_map[i].width==width)||  
			    (video_map[i].width<=0))){/*found an empty slot*/
      video_map[i].width=width;
      video_map[i].busy=1;
      break;
    }
    /* case 3*/
    if(!video_map[i].busy&&((video_map[i].width!=width)&&
			    (video_map[i].width>0))){
      clear_video_map(i);      
      video_map[i].width=width;
      video_map[i].busy=1;
      break;
    }
  }
  if(i>=WINDOW_LENGTH*WINDOW_WIDTH) {
    fprintf(stderr," can not support more than %i streams \n",i); 
    return i;
  }
  if(video_map[i].child_name[10]=='0') { /* in the empty state all the 
					  * child_name entries are set to 0 
					  * so there is no difference between 
					  * 10 or 20. however the first and
					  * second entry might be 0 for the 
					  * case of the 0 and 10 stream*/
    sprintf(video_map[i].child_name,".container.child%i",i);
    video_map[i].restart=0;
    count=i/WINDOW_WIDTH; 
    /* create a subwindow for WINDOW_WIDTH of video images */
    if(slave[count]==NULL) {
      char name[50];
      sprintf(name,".container.slave%i",count);
      if((slave[count]= 
	  Tk_CreateWindowFromPath(interp_, containerWindow, 
				    name, NULL))==NULL){
	fprintf(stderr," %s \n",interp_->result);
	perror("could not open child window");
	exit(0);
      }
      Tk_SetWindowBackground(slave[count],BACKGROUND);
      sprintf(command,"pack .container.slave%i ",count);
      Tcl_Eval(interp_,command);
      
    }
  }
  sprintf(video_map[i].command,"pack %s -side left -in .container.slave%i",
	  video_map[i].child_name,count);
  if(video_map[i].theImage) /*clear the found slot */
    XFillRectangle(video_map[i].theDisp,
		   Tk_WindowId(video_map[i].childWindow),
		   video_map[i].theGC_o,0,0,
		   video_map[i].theImage->width,
		   video_map[i].theImage->height+40);

  return i;
}



/**************************************************************************/
/*
 *close the X window created for XIL and bind the moue to the TK window 
 * return 0
 */
int ignore_XilStuff(int window,int width, int height)
{
  int i;
  for(i=0; i<WINDOW_LENGTH*WINDOW_WIDTH;i++) {
    if(Tk_WindowId(video_map[i].childWindow)==(Window)window)
      break;
  }
  if(i<WINDOW_LENGTH*WINDOW_WIDTH&&video_map[i].xilW>0){
    XFlush(video_map[i].theDisp);
    XSync(video_map[i].theDisp,TRUE);
    XDestroyWindow(video_map[i].theDisp,video_map[i].xilW);
  }
  video_map[i].xilW=0;
 return 0;
}  
/**************************************************************************/
/*
 * out of some strange reason xil refuses to write on windows created by TK
 * therefore we create another window with X over the TK window
 * this is a hack the compilcates the handling of YUV which uses xil for
 * decompression and display, but for now we have to live with it
 * return the X window id
 */

int do_XilStuff(int window,int width, int height)
{
  int i;
  XEvent event;
  Colormap colormap;
  XSetWindowAttributes attributes;
  unsigned long valuemask = 0;
  Window child_window;
  Tk_Window win;
  int screen_num;
  /* width/=s->scale_factor;
  height/=s->scale_factor;*/
  for(i=0; i<WINDOW_LENGTH*WINDOW_WIDTH;i++) {
    if(Tk_WindowId(video_map[i].childWindow)==(Window)window){
      win=video_map[i].childWindow;
      break;
    }
  }
  Tk_GeometryRequest(video_map[i].childWindow, width/s->scale_factor,
		     (height/s->scale_factor)+40);
  if(i<WINDOW_LENGTH*WINDOW_WIDTH){
    if(video_map[i].xilW>0)
      return video_map[i].xilW;
    screen_num = DefaultScreen(Tk_Display(win));
    colormap = XCreateColormap (Tk_Display(win),
				Tk_WindowId(win),
				Tk_Visual(win),
				AllocNone);
    attributes.background_pixel = FORGROUND;
    attributes.border_pixel = BACKGROUND;
    attributes.colormap = colormap;
    valuemask = CWBackPixel | CWBorderPixel | CWColormap ;
    child_window= XCreateWindow(Tk_Display(win),
				Tk_WindowId(win),
				0,0, width, height, 0,
				Tk_Depth(win),
				InputOutput, 
				Tk_Visual(win), 
				valuemask,&attributes); 
    /* list window events used. */
    XSelectInput(Tk_Display(win),child_window ,
		 ExposureMask | StructureNotifyMask);
    XMapWindow(Tk_Display(win),child_window);
    /* wait for the window to be mapped (an Expose event) */
    do
	XNextEvent(Tk_Display(win), &event);
    while (event.xany.type != Expose);
    return (int)child_window;
  }
  return 0;
}


/**************************************************************************/
/* 
 * Look for an empty video slot or create an new one for a video stream
 * and set the needed resources for such a slot for the calling stream
 * return 1 OK
 * return 0 no more entries left
 * return -1 ERROR
 */
int do_TkStuff(int* the_Image,int* the_Disp,int* the_GC,int* main_W,
	      int* the_Visual,int width,int height,int* shminfo,char *name,
	       int (*xil_stuff)())
{
  XColor* c;
  XGCValues v;
  u_long mask;
  int place;
  
  width*=s->scale_factor; /* scale_fatcor=1 -> no resizing */
  height*=s->scale_factor;

 /* Create the container window for all the sub images */
  if(first) {
    if(init_video_map()<0)
      return -1;
    first=0;
  }
  if((place=get_map(name,width))<WINDOW_LENGTH*WINDOW_WIDTH) {
    if(s->scale_factor==2) /* for the big picture case set found slot always to
			      free so when the active speaker changes we
			      can use it for the new one */
      video_map[place].busy=0;
    if(video_map[place].theImage==NULL) { /* the found slot has no resources
					   * so allocate the necessary stuff */
      video_map[place].name=name;
      if(video_map[place].restart==0) {
	video_map[place].restart=1;
	/* create a child window for the video images*/
	if((video_map[place].childWindow= 
	    Tk_CreateWindowFromPath(interp_, containerWindow, 
				    video_map[place].child_name, NULL))==NULL){
	  fprintf(stderr," %s \n",interp_->result);
	  perror("could not open child window");
	  return -1;
	}
	Tk_SetWindowBackground(video_map[place].childWindow,BACKGROUND);
	video_map[place].theDisp   = Tk_Display(video_map[place].childWindow);
	video_map[place].dispDEEP  = Tk_Depth(video_map[place].childWindow);
	video_map[place].theVisual = Tk_Visual(video_map[place].childWindow);
	
	/* initialize the graphics contents resources */
	c = Tk_GetColor(interp_, containerWindow, Tk_GetUid("gray50"));
	if (c == 0){
	  perror(" could not get init color ");
	  return -1;
	}
	/* create a GC for the video images */
	v.background = BACKGROUND;
	v.foreground = FORGROUND;
	mask = GCForeground|GCBackground;
	video_map[place].theGC =
	  Tk_GetGC(video_map[place].childWindow, mask, &v);
	
	/* create a GC for empty images */
	v.background = FORGROUND;
	v.foreground = BACKGROUND;
	video_map[place].theGC_o=
	  Tk_GetGC(video_map[place].childWindow,mask,&v);
	/* put the image on the screen */
	Tk_MapWindow(video_map[place].childWindow);
	
	/* handle the Exposure and destroy events for the created slot */
	Tk_CreateEventHandler(video_map[place].childWindow,
			      ResizeRequest|DestroyNotify|EnterNotify|LeaveNotify, 
			      Event_Proc, (ClientData) place);
      	Tcl_Eval(interp_,video_map[place].command);
	if((char *)Tk_WindowId(video_map[place].childWindow)==NULL)
	  perror(" still not mapped ");
      }
      /* handle the Exposure and destroy events for the created slot */
      Tk_CreateEventHandler(video_map[place].childWindow,
			    ResizeRequest|DestroyNotify|ExposureMask|StructureNotifyMask, 
			    Event_Proc, (ClientData) place);
      Tk_GeometryRequest(video_map[place].childWindow, width,height+40);


      /*allocate the image space for the video image */
      if(XShmQueryExtension(video_map[place].theDisp)) {
	if(place>=MAX_SHM) /* only the first MAX_SHM streams can use shared 
			    * memory segments*/
	  USE_SHM=0;
	else
	  USE_SHM=1;
      }
      if(USE_SHM)
	if((video_map[place].shminfo=(XShmSegmentInfo *)
	    malloc(sizeof(XShmSegmentInfo)))==NULL)
	  message_output("NOt Enough Memory",1);

      if(USE_SHM) { /* allocate a shared memory segment for the slot */
	if(!(video_map[place].theImage = 
	     XShmCreateImage(video_map[place].theDisp,
			     video_map[place].theVisual,
			     video_map[place].dispDEEP,ZPixmap, NULL,
			     video_map[place].shminfo,width,height))) {
	  FatalError("couldn't create Shared X image!");
	  return(-1);
	  
	}
	/* get a shared memory segment and attach it to the allocated image*/
	if((video_map[place].shminfo->shmid=
	    shmget(IPC_PRIVATE,video_map[place].theImage->bytes_per_line*
		   video_map[place].theImage->height,IPC_CREAT|0777))<0) {
	  perror("could not get shared memory using normal X server");
	  
	  /* as there seems to be no more shared memory segments to allocate
	   * retreat to normal slower X connection -this code seems not to
	   * be working correctly */
	  video_map[place].theImage = 
	    XCreateImage(video_map[place].theDisp, 
			 video_map[place].theVisual, 
			 video_map[place].dispDEEP, 
			 ZPixmap, 0,NULL,width,height,8,0);
	  if (!video_map[place].theImage) {
	    FatalError("couldn't create X image!");
	    return(-1);
	  }
	  video_map[place].theImage->data=
	    (u_char *)malloc((size_t)(height*
				   video_map[place].theImage->bytes_per_line));
	  if (!video_map[place].theImage->data) {
	    fprintf(stderr,"couldn't malloc image data of %i*%i bytes",height,
		      (int)video_map[place].theImage->bytes_per_line);
	    return(-1);
	  }
	  video_map[place].shminfo=NULL;	  
	  USE_SHM=0;
	}
	if(USE_SHM) {
	  video_map[place].theImage->obdata =(char*)video_map[place].shminfo;
	  video_map[place].shminfo->shmaddr=video_map[place].theImage->data= 
	    shmat(video_map[place].shminfo->shmid,0,0);
	  if(XShmAttach(video_map[place].theDisp,video_map[place].shminfo)
	     ==0){
	    perror(" Can not XShmAttach");
	    return(-1);
	  }
	  video_map[place].shminfo->readOnly=False;
	}
      }
      else { /* allocate a normal Image and buffer  space for the stream with 
	      * no shared memory segments */
	video_map[place].theImage = 
	  XCreateImage(video_map[place].theDisp, 
		       video_map[place].theVisual, 
		       video_map[place].dispDEEP, 
		       ZPixmap, 0,NULL,width,height,8,0);
	if (!video_map[place].theImage) {
	  FatalError("couldn't create X image!");
	  return(-1);
	}
	video_map[place].theImage->data = 
	  (u_char *) malloc((size_t) (height*
				   video_map[place].theImage->bytes_per_line));
	if (!video_map[place].theImage->data) {
	  fprintf(stderr,"couldn't malloc image data of %i*%i bytes",height,
		  (int)video_map[place].theImage->bytes_per_line);
	  return(-1);
	}
      }
    }
    else {/*there is an empty slot in the window so use it for the new stream*/
      XFillRectangle(video_map[place].theDisp,
		     Tk_WindowId(video_map[place].childWindow),
		     video_map[place].theGC_o,0,0,
		     video_map[place].theImage->width,
		     video_map[place].theImage->height+40);
      video_map[place].name=name;
    }
  }
  else {
    fprintf(stderr," currently only %i video slots are supported \n",
	    WINDOW_WIDTH*WINDOW_LENGTH);
    return 0;
  }
  Tk_GeometryRequest(video_map[place].childWindow, width,height+40);
  { /* bind the mouse click to this window */
    char command[50];
    sprintf(command,"bind %s <1> {resize %i}",
	    video_map[place].child_name,place);
    Tcl_Eval(interp_,command);
  }
  video_map[place].xilW =xil_stuff(Tk_WindowId(video_map[place].childWindow),
				    width,height);
  /* export the allocated resources */
  *the_Disp    =  Tk_Display(video_map[place].childWindow);
  *the_Image   =  video_map[place].theImage ;        
  *the_Visual  =  video_map[place].theVisual;        
  *the_GC      =  video_map[place].theGC    ;        
  
  *main_W       =  (((int)video_map[place].xilW!=0) ? video_map[place].xilW :
		    Tk_WindowId(video_map[place].childWindow));
  if(USE_SHM)
    *shminfo   =  video_map[place].shminfo;        
  uname(&u_name);
  return(1);  
}


/***********************************/
void drawWindow(int x,int y,int w,int h,char *the_Image,char* the_Disp,
		char* the_GC,char* main_W,char* name,char* shminfo)
{ 
  XImage* theImage=(XImage*)the_Image;
  Display* theDisp=(Display*)the_Disp;
  GC theGC=(GC)the_GC;
  Window mainW=(Window)main_W;
  w*=s->scale_factor;
  h*=s->scale_factor;


  if (theImage!=NULL) {
    if(shminfo!=NULL) { /* using shared memory*/
      XShmPutImage(theDisp,mainW,theGC,theImage,x,y,x,y,(u_int)w, 
	(u_int) h,0);
      XSynchronize(theDisp,True);
    }
    else {
      XPutImage(theDisp,mainW,theGC,theImage,x,y,x,y, (u_int) w, 
		(u_int) h);
    }
    /* display the name of the sender */
    if(name!=NULL)
      XDrawString(theDisp,mainW,theGC,(w/2)-strlen(name),h+20,name,strlen(name));
  }
  else
    fprintf(stderr,"Tried to DrawWindow when theImage was NULL\n ");

}

/***********************/
/* 
 * after a resize clear the videomap and init videomap[0] for the active window
 */

void clear_TkAll() 
{
  int i;
  for(i=0; i<WINDOW_LENGTH*WINDOW_WIDTH;i++) {
    if(video_map[i].theImage!=NULL) {
      /* stop the event handling */
      Tk_DeleteEventHandler(video_map[i].childWindow,
                            ResizeRequest|DestroyNotify|ExposureMask|StructureNotifyMask,
                            Event_Proc, (ClientData) i);
      video_map[i].busy=0;  
      Tk_GeometryRequest(video_map[i].childWindow, 1,1);
    }    
  }
}

/**************************************************************/
/* clear the videomap[0] and set its TheImage to NULL
 */

void clear_video_map(int window)
{
  /* prepare videp_map[window] to contain the active stream */ 
  if(video_map[window].theImage!=NULL) {
    /* stop the event handling */
    Tk_DeleteEventHandler(video_map[window].childWindow,
			  ResizeRequest|DestroyNotify|ExposureMask|StructureNotifyMask,
			  Event_Proc, (ClientData) window);
    /* free the allocate buffer and shared memort segments */
    if(video_map[window].shminfo!=NULL) {
      XShmDetach(video_map[window].theDisp, video_map[window].shminfo);
      XDestroyImage(video_map[window].theImage);
      if (shmdt(video_map[window].shminfo->shmaddr) < 0)
	perror("nevit: shmdt");
    }
    else {
      free(video_map[window].theImage->data);
      video_map[window].theImage->data=NULL;
      XDestroyImage(video_map[window].theImage);
    }
    video_map[window].restart=1;
    video_map[window].width=0; /* reinitialize the width */
    video_map[window].theImage=NULL;
  }
}

/***********************/
/* 
 * do the final clean up of the video stuff by freeing all allocated resources
 */

void end_TkAll(int flag) 
{
  int i;
  for(i=0; i<WINDOW_LENGTH*WINDOW_WIDTH;i++) {
    if(video_map[i].theImage!=NULL) {
      /* stop the event handling */
      Tk_DeleteEventHandler(video_map[i].childWindow,
                            ResizeRequest|DestroyNotify|ExposureMask|StructureNotifyMask,
                            Event_Proc, (ClientData) i);
    /* free the allocate buffer and shared memort segments */
      if(video_map[i].shminfo!=NULL) {
	XShmDetach(video_map[i].theDisp, video_map[i].shminfo);
	XDestroyImage(video_map[i].theImage);
	if (shmdt(video_map[i].shminfo->shmaddr) < 0)
	  perror("nevit: shmdt");
      }
      else {
	free(video_map[i].theImage->data);
	video_map[i].theImage->data=NULL;
	XDestroyImage(video_map[i].theImage);
      }
      video_map[i].busy=0;  
      Tk_GeometryRequest(video_map[i].childWindow, 1,1);
      if(!flag){ /* free the GC only if the programm is ending flag=0 */
	Tk_FreeGC(video_map[i].theDisp,video_map[i].theGC);
	Tk_FreeGC(video_map[i].theDisp,video_map[i].theGC_o);
      }
      video_map[i].theImage=NULL;
      video_map[i].name=NULL;
      video_map[i].width=0;
    }    
  }
  /* free the allocated color map*/
  free_colors();
}
/***********************/
/*
 * set the window slot of name as free to be used by other streams
 * and color it with the background color 
 */
int end_TkStuff(char* name)
{
  int i=0;
  /* search for the slot diplaying the data of name */
  for(i=0; i<WINDOW_LENGTH*WINDOW_WIDTH;i++) {
    if((video_map[i].name!=NULL)&&!strcmp(name,video_map[i].name)) {
      break;
    }
  }
  if(i>=WINDOW_LENGTH*WINDOW_WIDTH) {
    perror("Something worng with closing the window");
    return -1;
  }
  
  /* clear the slot and write "no entry" on it */
  if(video_map[i].theImage!=NULL){
    XFillRectangle(video_map[i].theDisp,Tk_WindowId(video_map[i].childWindow),
		   video_map[i].theGC_o,0,0,
		   video_map[i].theImage->width,
		   video_map[i].theImage->height+40);
    video_map[i].busy=0;
    XDrawString(video_map[i].theDisp,Tk_WindowId(video_map[i].childWindow),
		video_map[i].theGC,(video_map[i].theImage->width/2)-8,
		video_map[i].theImage->height+20,
		"   no  entry    ",16);
  }  
  if(video_map[i].xilW>0){
    XFlush(video_map[i].theDisp);
    XSync(video_map[i].theDisp, TRUE);
    XDestroyWindow(video_map[i].theDisp,video_map[i].xilW );
    video_map[i].xilW=0;
  }
  for(i=0;i<20;i++)
      video_map[i].child_name[i]     =   '0';
  
  return 1;
}
/***********************************/

int FatalError (char * identifier)
{
  fprintf(stderr, " %s\n", identifier);
  return(-1);
}









