#ifndef __mbuf_h
#define __mbuf_h
#include "types.h"
#include "audio_descr.h"

typedef struct mbuf {
  struct mbuf *m_next; /* next buffer in chain */
  unsigned long m_off; /* offset to data (from head of mbuf) */
  short m_len;         /* amount of data in this mbuf */
  short m_type;        /* mbuf type (0 == free) */
  short m_ref;         /* reference count */
  u_int32 m_timestamp; /* timestamp of packet */
  audio_descr_t m_ad;  /* audio content of this mbuf */
  u_char mun_dat[1];   /* data storage */
} mbuf_t;

#define MMINOFF(m)  ((char *)m->mun_dat - (char *)m)
#define MT_FREE (-1)    /* unallocated */

extern mbuf_t *m_get(char *);
extern mbuf_t *m_copy(mbuf_t *m);
extern int     m_create(int count, size_t alloc);
extern void    m_free(mbuf_t *m, char *);
extern void    m_freem(mbuf_t *mh, char *);

/* Current start of data in mbuf. */
#define mtod(x,t)       ((t)((long)(x) + (x)->m_off))

/* Append point for data in mbuf. */
#define mtod_end(x,t)   ((t)((long)(x) + (x)->m_off + (x)->m_len))

/* Remove header of length 'x'. */
#define m_adj(m,x)      ((m)->m_off += (x), (m)->m_len -= (x))

/* Adjust mbuf pointers to prepend 'x' bytes. */
#define mprepend(m,x)   ((m)->m_off -= (x), (m)->m_len += (x))
#endif
