#ifndef _sync_h_
#define _sync_h_
/*
* Synchronization state; time is measured in sample ticks (i.e., at
* the sampling frequency), unaffected by the number of audio channels. 
* For example, the sample clock for CD-quality stereo ticks at 48 kHz.
*
* $Id: sync.h,v 1.1.1.1 1997/11/12 17:16:33 dor Exp $
*/
typedef struct {
  struct {
    u_int32 ts;        /* timestamp (samples) */
    u_int32 seq;       /* sequence number */
  } first, this, prior;  /* first, current and previous packet */

  /* extended sequence number */ 
  u_int64 cycles;      /* complete sequence number cycles */
  u_int32 max_seq;     /* highest sequence number */
  u_int32 bad_seq;     /* out-of-line sequence number */

  u_int32 t;           /* current time, measured in samples */
  u_int32 p;           /* current playout point (samples) */
  u_int32 slack;       /* time left over until playout (samples) */
  u_int32 last_audio;  /* furthest audio (samples) */
  u_int32 delay;       /* offset between ts and playout pointer p */
  u_int32 d_est;       /* estimated clock offset */
  double d_avg;        /* estimate of average slack */
  double d_var;        /* estimate of slack variation */
} sync_t;
#endif
