#
# fileselect_frame.tcl --
# simple file selector in a frame.
#
# Layout:
#
#  file:                  +----+
#  ____________________   | OK |
#                         +----+
#
#  +------------------+    Cancel
#  | ..               |S
#  | file1            |c
#  | file2            |r
#  |                  |b
#  | filen            |a
#  |                  |r
#  +------------------+
#  current-directory
#
# Copyright 1993 Regents of the University of California
# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and without
# fee is hereby granted, provided that this copyright
# notice appears in all copies.  The University of California
# makes no representations about the suitability of this
# software for any purpose.  It is provided "as is" without
# express or implied warranty.
#
# $Id: fileselect_f.tcl,v 1.1.1.1 1997/11/12 17:16:27 dor Exp $
#


# Names starting with "fileselect" are reserved by this module;
# no other names used.
# Modified by Henning Schulzrinne, 14-Nov-1993

# Use the "option" command for further configuration

#option add *Listbox*font \
#    "-*-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1" startupFile
#option add *Entry*font \
#    "-*-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1" startupFile
#option add *Label*font \
#    "-*-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1" startupFile


# This is the default proc  called when "OK" is pressed;
# to indicate yours, give it as the first arg to "fileselect"

proc fileselect.default.ok_cmd {f} {
  puts stderr "[pid] selected file $f"
}

proc fileselect.default.cancel_cmd {f} {
}


#
# proc that creates the file selector box
# (external interface)
#
proc fileselect {
    {w .fileSelectWindow}
    {ok_cmd fileselect.default.ok_cmd} 
    {cancel_cmd *}
    {purpose "Open file:"} 
    {dir .}
    {pattern *}
    {ok Ok}} {

  # destroy duplicates
  catch {destroy $w}

  frame $w

  # path independent names for the widgets
  global fileselect_entry fileselect_list \
    fileselect_ok fileselect_cancel fileselect_dirlabel \
    fileselect_pattern fileselect_dir

  set fileselect_pattern $pattern
  set fileselect_entry $w.file.eframe.entry
  set fileselect_list $w.file.sframe.list
  set fileselect_scroll $w.file.sframe.scroll
  set fileselect_ok $w.bframe.okframe.ok
  set fileselect_cancel $w.bframe.cancel
  set fileselect_dirlabel $w.file.dirlabel

  # if 'dir' doesn't exist, show current directory
  if {[file isdirectory $dir] == 0} {
    tkerror "$dir doesn't exist - using current directory"
    set dir "."
  }

  if {$dir == "."} {
    set fileselect_dir [exec pwd]
  } else {
    set fileselect_dir $dir
  }

  # widgets
  frame $w.file -bd 10 
  frame $w.bframe -bd 10
  pack $w.file -side left -fill y
  pack $w.bframe -side left -expand true -anchor n

  frame $w.file.eframe
  frame $w.file.sframe
  label $w.file.dirlabel -anchor e -width 24 -text $fileselect_dir

  pack $w.file.eframe   -side top -anchor w
  pack $w.file.sframe   -side top -fill x
  pack $w.file.dirlabel -side top -anchor w

  label $w.file.eframe.label -anchor w -width 24 -text $purpose
  entry $w.file.eframe.entry -relief sunken 
  entry $w.file.eframe.pattern -relief sunken -textvariable fileselect_pattern
  label $w.file.eframe.plabel -anchor w -text Pattern:

  pack $w.file.eframe.label   -side top    -expand true -anchor w
  pack $w.file.eframe.entry   -side top    -fill x -anchor w
  pack $w.file.eframe.pattern -side bottom -fill x -anchor e
  pack $w.file.eframe.plabel  -side bottom -fill x -anchor e

  scrollbar $w.file.sframe.yscroll -relief sunken \
	 -command "$w.file.sframe.list yview"
  listbox $w.file.sframe.list -relief sunken \
	-yscroll "$w.file.sframe.yscroll set" 

  pack $w.file.sframe.yscroll -side right -fill y
  pack $w.file.sframe.list    -side left -fill y 

  # buttons
  frame $w.bframe.okframe -borderwidth 2 -relief sunken
 
  button $w.bframe.okframe.ok -text $ok -relief raised -padx 2 \
        -command "fileselect.ok.cmd $w $ok_cmd"

  button $w.bframe.cancel -text cancel -relief raised -padx 10 \
        -command "fileselect.cancel.cmd $w $cancel_cmd"
  pack $w.bframe.okframe.ok -padx 10 -pady 10

  pack $w.bframe.okframe -expand true -padx 20 -pady 20
  pack $w.bframe.cancel -side top

  # Fill the listbox with a list of all the files in the directory (run
  # the "ls" command to get that information).
  fileselect.renew

  # Set up bindings for the browser.
  bind $fileselect_entry <Return> {
    eval $fileselect_ok invoke
  }
  bind $fileselect_entry <Control-c> {eval $fileselect_cancel invoke}
  bind $w.file.eframe.pattern <Return> {fileselect.renew }

  bind $w <Control-c> {eval $fileselect_cancel invoke}
  bind $w <Return> {eval $fileselect_ok invoke}

  bind $fileselect_list <ButtonRelease-1> {
    %W select anchor [%W nearest %y]
    %W select set anchor [%W nearest %y]
	eval $fileselect_entry delete 0 end
	eval $fileselect_entry insert 0 [%W get [%W nearest %y]]
  }

  bind $fileselect_list <Double-ButtonPress-1> {
	eval $fileselect_ok invoke
  }

  bind $fileselect_list <Return> {
    %W select anchor [%W nearest %y]
    %W select set anchor [%W nearest %y]
    eval $fileselect_entry delete 0 end
    eval $fileselect_entry insert 0 [%W get [%W nearest %y]]
	eval $fileselect_ok invoke
  }

  return $w
}


# auxiliary button procedures

proc fileselect.cancel.cmd {w cmd {arg ""}} {
  if {$cmd == "*"} {
    destroy $w
  } else {
    eval $cmd $arg
  }
}

#
# show directory 'dir'
#
proc fileselect.renew {} {
  global fileselect_entry fileselect_dirlabel fileselect_list \
    fileselect_pattern fileselect_dir

  eval $fileselect_dirlabel configure -text $fileselect_dir
  eval $fileselect_entry delete 0 end
  eval $fileselect_list delete 0 end
  # need to expand tilde for ls
  set fileselect_dir [glob $fileselect_dir]

  # if 'dir' doesn't exist, show current directory
  if {[file isdirectory $fileselect_dir] == 0} {
    tkerror "$fileselect_dir doesn't exist - using current directory"
    set fileselect_dir [exec pwd]
  }

  foreach i [exec ls -a $fileselect_dir] {
    if {([string compare $i "."] != 0 && \
        [string match $fileselect_pattern $i]) || \
        [string compare $i ".."] == 0} {
      eval $fileselect_list insert end $i
    }
  }
}

#
# Command invoked when 'Ok' button is pressed
#
proc fileselect.ok.cmd {w cmd {arg ""}} {
  global fileselect_entry fileselect_dirlabel fileselect_list \
  fileselect_dir
  set selected [$fileselect_entry get]

  # if $selected starts with /, it's an absolute filename,
  # otherwise, it's appended to $fileselect_dir
  if {$selected == ".."} {
    set fileselect_dir [file dirname $fileselect_dir]
  } elseif {[string index $selected 0] == "/"} {
    set fileselect_dir $selected
  } elseif {[file isdirectory $fileselect_dir/$selected]} {
    set fileselect_dir $fileselect_dir/$selected
  } else {
    eval $cmd $arg $fileselect_dir/$selected
    return
  }
  fileselect.renew
}
