/*
convert  --  convert to local (workstation) audio format

Copyright 1994 by AT&T Bell Laboratories; all rights reserved
*/

#include "nevot.h"
#include "audio_generic.h"
#include "g711.h"
#include "g721.h"
#include "g723.h"
#include "dvi_adpcm.h"
#include "gsm.h"
#include "lpc.h"
#include "l16.h"
#include "channels.h"
#include "control.h"
#include "member.h"
#include <ansi.h>



/* order must be same as in audio_descr.h */
codec_t codec[] = {
/* AE_      name           bps     bpm init            encode            decode        */
/* PCMU */ {"G.711 u-law", 1.0,     1, g711_init,      g711_encode,      g711_decode},
/* PCMA */ {"G.711 A-law", 1.0,     1, g711_init,      g711_encode,      g711_decode},
/* G721 */ {"G.721",       0.5,     1, G721_init,      G721_encode,      G721_decode},
/* DVI4 */ {"DVI ADPCM",   0.5,     1, dvi_adpcm_init, dvi_adpcm_encode, dvi_adpcm_decode},
/* G723 */ {"G.723",       0.375,   1, G723_init,      G723_encode,      G723_decode},
/* GSM  */ /*{"GSM",         0.20625, 4, GSM_init,       GSM_encode,       GSM_decode},*/
/* 1016 */ {"FedStd 1016", 0.075,   4, 0,              0,                0},
/* change LPC bps once ``real'' LPC is being used */
/* LPC  */ {"LPC-10",      0.0777,  4, lpc_init,       lpc_encode,       lpc_decode},
/* L8   */ {"linear 8",    1.0,     1, 0,              0,                0},
/* L16  */ {"linear 16",   2.0,     1, l16_init,       l16_encode,       l16_decode},
/* L24  */ {"linear 24",   4.0,     1, 0,              0,                0},
/* L32  */ {"linear 32",   4.0,     1, 0,              0,                0},
/* G728 */ {"G.728",       0.5,     1, 0,              0,                0}, 
/* TSP0 */ {"true speech", 1.0,     1, 0,              0,                0}, 
/*      */ {"unknown",     0.0,     0, 0,              0,                0}
};

/*
* Return the number of samples contained in packet of 'bytes' bytes.
* Take stereo and single encoding header into account.
*/
int bytes2samples(audio_descr_t *adp, int bytes)
{
  /* The SPARC takes three times as long to divide by 1 as by
     any other integer; thus we check for this case. */
  /* need to take channels into account */
  if (adp->channels > 1) bytes /= (int)adp->channels;

  /* encoding translation, without regard to sampling rate */
  switch(adp->encoding) {
  case AE_PCMU: /* framecount equals byte count */
  case AE_PCMA:
    break;
  case AE_DVI4:
    bytes -= 4;       /* state header */
  case AE_G721:       /* 1/2 */
    bytes <<= 1; 
    break;
  case AE_G723:       /* 8/3 */
    bytes *= 2.666667;
    break;
  case AE_GSM:        /* 33 bytes -> 160 samples */
    bytes *= 4.8485;
    break;
  case AE_1016:       /* XXX */
    break;
  case AE_LPC:        /* XXX (estimate only) */
    bytes = (bytes / sizeof(lpc_t)) * 180;
    break;
  case AE_L8:         /* frames == bytes */
    break;
  case AE_L16:        /* frame count is half byte count */
    bytes >>= 1;
    break;
  case AE_L24:        /* divide by three */
    bytes *= 0.33333333;
    break;
  case AE_L32:        /* divide by four */
    bytes >>= 2;
    break;
  default:
    break;
  }
  return bytes;
} /* bytes2samples */
