#include "Audio.h"
#include <string.h>

/*
* Translate error number 'e' into error message. If no message is known,
* return printable value of 'e'.
* XXX Insert internationalization/message library here.
*/
char *Audio_strerror(Audio_errno_t e)
{
  extern char *sys_errlist[];
  extern int errno;

  static char msg[256];
  switch(e) {
    case Audio_ECNOMCAST:
      return "not compiled for multicast";
    case Audio_EHNOMCAST:
      return "host does not support multicast";
    case Audio_EMIXED:
      return "mixed multicast/unicast";
    case Audio_ESELF:
      return "cannot connect to self";
    case Audio_ENOHOST:
      return "no such host";
    case Audio_EAUDIO_OPEN:
      return "cannot open audio device";
    case Audio_EAUDIO_BUSY:
      return "audio device in use";
    case Audio_EAUDIO_ACCES:
      return "access to audio device denied";
    case Audio_EAUDIO_INTR:
      return "audio system call interrupted";
    case Audio_EAUDIO_IO:
      return "hangup or error during audio open";
    case Audio_EAUDIO_NXIO:
      return "audio device does not exist";
    case Audio_EINVAL:
      return "invalid parameter";
    case Audio_ERATE:
      return "cannot convert sampling rate";
    case Audio_ECHANNELS:
      return "cannot convert channel";
    case Audio_EENC:
      return "cannot convert encoding";
    case Audio_EODDPORT:
      return "cannot use odd port number";
    case Audio_ECANTREAD:
      return "cannot read file";
    case Audio_EIHX:
      return "invalid header extension";
    case Audio_EIAUDIO:
      return "invalid audio type";
    case Audio_EICONFID:
      return "invalid conference id";
    case Audio_EISSRC:
      return "invalid source identifier";
    case Audio_EIPAD:
      return "invalid padding";
    case Audio_EITYPE:
      return "invalid control packet type";
    default:
      strcpy(msg, sys_errlist[errno]);
      return msg;
  }
} /* Audio_strerror */
