/*
tcl  --  set system parameters through Tcl.

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include <tcl.h>
#include "util.h"           /* itoa() */
#include "audio_generic.h"  /* audio_...() */
#include "list.h"           /* list_to_index() */
#include "nevot.h"          /* nevot.rx */
#include <math.h>           /* exp() */
#include <stdlib.h>         /* atoi() */
#include <string.h>         /* strcmp() */
#include <ansi.h>



#include "rx.h"             /* RX_t */
extern struct RX_t RX;
extern void audio_vu(int, float);
extern void audio_vu_max(int, float);

/*
* Tcl callback to configure receiver parameters. This routine is also
* called directly if the sampling rate changes so that dependent
* parameters can be updated.
*/
static char *rx_trace(ClientData client, Tcl_Interp *interp, 
  char *name1, char *name2, int flags)
{
  char *value = Tcl_GetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
  double v;

  if (!value) return 0;  /* if variable isn't defined yet */
  v = atof(value);
  if (!RX.pm) RX.pm = peakmeter_init(1, audio_vu, audio_vu_max);

  if (!name2) {
    return "is an array";
  }
  else if (strcmp(name2, "hard_timeout") == 0) {
    RX.timeout[1] = v;
  }
  else if (strcmp(name2, "soft_timeout") == 0) {
    RX.timeout[0] = v;
  }
  else {
    return "undefined rx parameter";
  }
  return NULL;
} /* rx_trace */


/*
* Initalizes receiver and traceback for "rx" variable.
*/
void rx_config(Tcl_Interp *interp)
{
  Tcl_TraceVar(interp, "rx", TCL_TRACE_WRITES, rx_trace, 0);
  Tcl_TraceVar(interp, "vu", TCL_TRACE_WRITES | TCL_GLOBAL_ONLY, peakmeter_trace,
    (ClientData)RX.pm);
} /* rx_init */

#include "tx.h"             /* TX_t */
extern struct TX_t TX;

/*
* Configure audio transmission. Also called when audio parameters
* change; we adapt packetization accordingly.
*/
static char *tx_trace(ClientData client, Tcl_Interp *interp, 
  char *name1, char *name2, int flags)
{
  extern void tx_packet_init(void);
  char *value = Tcl_GetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
  int v;
  double d;

  if (!value) return 0;  /* variable not yet defined (if direct call) */
  v = atoi(value);
  d = atof(value);

  if (!name2) {
    return "is an array";
  }
  else if (strcmp(name2, "after") == 0) {
    TX.spurt[1] = v;  /* packetization intervals */
  }
  else if (strcmp(name2, "before") == 0) {
    TX.spurt[0] = v;   /* packetization intervals */
  }
  else if (strcmp(name2, "release_audio") == 0) {
    TX.release = d;
  }
  else if (strcmp(name2, "monitor") == 0) {
    TX.in.monitor = (v != 0);
    TX.monitor = TX.f.f ? TX.f.monitor || TX.in.monitor : TX.in.monitor;
  }
  else {
    return "undefined variable";
  }
  return NULL;
} /* tx_trace */

/*
* Configure audio transmission (file, chunk).
*/
void tx_config(Tcl_Interp *interp)
{
  Tcl_TraceVar(interp, "vu",  TCL_TRACE_WRITES | TCL_GLOBAL_ONLY,
    peakmeter_trace, (ClientData)TX.pm);
  Tcl_TraceVar(interp, "tx",  TCL_TRACE_WRITES, tx_trace,  0);
  Tcl_TraceVar(interp, "sd",  TCL_TRACE_WRITES, sd_trace,  (ClientData)TX.sd);
  Tcl_TraceVar(interp, "agc", TCL_TRACE_WRITES, agc_trace, (ClientData)TX.agc);
} /* tx_config */
