#
# This file implements the pmm protcol for pmm-rsvp.
# 
#
# ----------------------------------------------------------------------
#
# initial value of activity inirtval reports is set to 2000, the actual value
# to be used will then be delivered by isc
# 
set update_int 2000



proc session args {
        global V
        eval $V(session) $args
}

proc pmm_eval cmd {
    global pmm sessions
    set pmm(sn) [lindex $cmd 0]
    if {[string length $cmd] > 0} {
	if {[lsearch {proc exec rename} [lindex $cmd 0]] >= 0} {
	    puts stderr "[pid] $pmm(tool): forbidden command: $cmd"
	} elseif {[string match "session *" $cmd]}  {
	    if {[string match "*/reservation/*" [lindex $cmd 1]]} {
		set pmm(sn) [lindex $cmd 1]
		if {$pmm(debug) > 0} { puts "$cmd pmm_$pmm(sn)" } 
		eval "proc pmm_$pmm(sn) {c args} \
                                           {pmm_session $pmm(sn) \$c \$args}"
		pmm_send "$pmm(sn) created"
		set myself [idstring %u@%d]
		#				pmm_send_ "hello $myself"
		set host [get_host $pmm(sn)]
		set port [get_port $pmm(sn)]
		set sessions [ladd $sessions $pmm(sn)]
		rsvp rapi_session $host $port
	    }
	} else {
	    if [catch {eval pmm_$cmd} msg] {
	    }
	}
    }
}

#
# Receive UDP packet
#
proc pmm_read_udp { sock} {
	pmm_eval [gets $sock]
}

#
# Open UDP connection 
#
proc pmm_init_udp {controller} {
	global pmm c

	set host [lindex $controller 0]
	set port [lindex $controller 1]
	set c [udp $host $port]
	Tcl_SetSockBuffer $c
	fconfigure $c -ttl 0  
	fconfigure $c -buffering none
	fconfigure $c -member 1
	fileevent $c readable {pmm_read_udp $c}

	proc pmm_send {msg} {
	    global pmm c cdest cfile prog_name
	    
	    if {$pmm(debug) > 0} {puts "-> $msg"}
	    puts -nonewline $c $msg 
	}
}

# Like pmm_send, but prefix our session name.
proc pmm_send_ msg {
	global pmm
	if [info exists pmm(sn)] {
		pmm_send "$pmm(sn) $msg"
	} 

}

#
# Invoked when a pmm session command was received.
#
proc pmm_session {st c args} {
    global pmm V sending_state path_state reserving_state update_int

    if {[string match source* $c]} {
	set source [string range $c 6 end]
	set c "source"
    }
    if {$pmm(debug) > 0} {puts " [info nameofexecutable]-> $st $c $args"}
    switch $c {
	open {
	    pmm_send "$st opened"
	    set myself [idstring %u@%d]
	    pmm_resend_path
	}
	update_interval {
	    set update_int [lindex $args 0]
	}
	local {
	    if {![string match "*{}*" $args]} {
		set host [get_host $st]
		set port [get_port $st]
		rsvp set_local $host $port $args
		
	    }
	}
	hello {
	    set cname $args
	    pmm_send "$st opened"
	}
	reserve {
	    regsub -all \{ $args "" args
	    regsub -all \} $args "" args
	    if { $args != ""} {
		set host [get_host $st]
		set port [get_port $st]
		if {[llength [split $args]]==1} {
		    rsvp reserve $host $port $args 
		    
		} else {
		    set value  [lindex [split $args] 0]
		    set source [lindex [split $args] 1]
		    if { $source == 0 } return;
		    rsvp set_reservation $host $port $value $source
		    rsvp reserve $host $port $args 
		}
	    }
	}
	source {
	    if {$args != "{{}}"} {
		set host [get_host $st]
		set port [get_port $st]
		rsvp set_reservation $host $port $args $source 
	    }
	}
	path {
	    set host [get_host $st]
	    set port [get_port $st]
	    if {($args == 1)} {
		rsvp rapi_sender $host $port $args
		set path_state 0
	    } elseif {($args == 0)} {
		rsvp rapi_sender $host $port $args
		set path_state 0
	    }
	}
	flowspec {
	    regsub -all "\{" $args "" args_new
	    regsub -all "\}" $args_new "" args_new
	   # if { [llength [split $args_new " "] ]< 9 } return
	    
	    set src [lindex [split $args_new " "] 0]
	    set type [lindex [split $args_new " "] 1]
	    set b [lindex [split $args_new " "] 2]
	    set r [lindex [split $args_new " "] 3]
	    set p [lindex [split $args_new " "] 4]
	    set m [lindex [split $args_new " "] 5]
	    set M [lindex [split $args_new " "] 6]
	    set R [lindex [split $args_new " "] 7]
	    set S [lindex [split $args_new " "] 8]
	    set host [get_host $st]
	    set port [get_port $st]
	    rsvp rapi_flowspec $host $port $type $b $r $p $m $M $R $S $src
	}
	tspec {
	    regsub -all "\{" $args "" args_new
	    regsub -all "\}" $args_new "" args_new
	    set type [lindex [split $args_new " "] 0]
	    set b [lindex [split $args_new " "] 1]
	    set r [lindex [split $args_new " "] 2]
	    set p [lindex [split $args_new " "] 3]
	    set m [lindex [split $args_new " "] 4]
	    set M [lindex [split $args_new " "] 5]
	    set host [get_host $st]
	    set port [get_port $st]
	    rsvp rapi_tspec $host $port $type $b $r $p $m $M
	}
	filtertype {
	    set host [get_host $st]
	    set port [get_port $st]
	    rsvp rapi_filtertype $host $port $args
	    
	}
	send {
	    set sending_state $args
	}
	close {
	    set host [get_host $st]
	    set port [get_port $st]
	    rsvp rapi_release $host $port
	    pmm_adios

	}
	bye {
	    pmm_adios
	}
    }
}

# close a session
proc pmm_adios {} {
    global pmm sessions
    set myself [idstring %u@%d]
    pmm_send_ "$myself bye"
    pmm_send_ "closed"
    set sessions [lremove $sessions $pmm(sn)]
    if { [llength $sessions] == 0 } {
	adios_exit
    }
}

# Close the application and sends bye meassages to isc
proc adios_exit {} {
	global sessions
        set myself [idstring %u@%d]
        foreach i $sessions {
                set host [get_host_normal $i]
                set port [get_port $i]
                pmm_send "$i closed"
        }
	#rsvp close
	exit
}

# Invoked while starting the application	
proc pmm_init {{mode ""} {debug 0}} {
    global pmm V path_msg tspec_msg sending_state path_state reserving_state sessions MODE update_int
    
    set MODE $mode
    set sending_state 0
    set path_state 0
    set reserving_state 0
    set sessions ""
    set path_msg ""
    set tspec_msg ""

    build.window$MODE .
    
    set pmm(debug) $debug
    set pmm(controller) "225.2.0.1 5010"
    set pmm(tool) rsvp-pmm
    pmm_init_udp $pmm(controller)
    set myself [idstring %u@%d]
    
    set pmm(tool) rsvp-pmm
    pmm_send "agent $pmm(tool) attach [pid]"

    # start activity messages
    # schedule the activity messages
    after $update_int update_active 
     
    #rsvp open
}

#
# signal the activity of the tool
#
proc update_active { } {
    global sessions update_int

    foreach e $sessions {
	if { $e != "" } {
	    pmm_send "$e active"
	}
    }
    if { $update_int > 0 } {
       after $update_int update_active
    }
}
# return the host name in dotted notation
proc get_host {st} {
    set host [lindex [split [lindex [split $st "/"] \
				 [expr [llength [split $st "/"]] -1]] ":" ] 0 ]
    set host [host2ip $host]
    return $host
}
#return the host name as was delivered by pmm
proc get_host_normal {st} {
    set host [lindex [split [lindex [split $st "/"] \
				 [expr [llength [split $st "/"]] -1]] ":" ] 0 ]
    return $host
}

proc get_port {st} {	
    set port [lindex [split [lindex [split $st "/"] [expr [llength [split $st "/"]] -1]] ":" ] 1] 
	return $port
}



#Remove element 'e' from variable 'v'.
proc lremove {v e} {
 
  set i [lsearch $v $e]
  if {$i >= 0} {
    set v [lreplace $v $i $i]
  }
  return $v
}
 
#
# Add 'e' to 'v' only if not in list already. Sort list.
#
proc ladd {v e} {

  set i [lsearch $v $e]
  if {$i < 0} {
    lappend v $e
  }
  set v [lsort $v]
  return $v
}
#
# resend all the received path messages
#
proc pmm_resend_path { } {
    global sessions path_msg tspec_msg

    foreach i $path_msg {
	if { $path_msg  != "" } {
	    pmm_send "$ses [lindex $path_msg $i]"
	    pmm_send "$ses [lindex $tspec_msg $i]"
	    set l [expr $l+1]
	}
    }
}
proc convert_2_k { value } { 
    
    if {![regexp {[0-9]} $value]} {
	set value ""
    }
    if {[regexp M $value]} {
	regsub -all M $value "" value;set value [expr $value * 1000000]
	#regsub -all M $value "" value;set value $value
    } else {
	if {[regexp K $value]} {
	    regsub -all K $value "" value;set value [expr $value * 1000]
	} else {
	    # akp divided all arriving tspec by 1000, 
	    # dor updated this value to stay as it is 
	   # set value [expr $value / 1000]
	    set value $value
	}
    }
    return $value
}


#
# analyze the received tspec, convert it to the format used in isc  
# and resend it over the pmm bus
#
proc pmm_tspec  {tspec} {
    regsub -all {[a-z]|=|\(|\)|\[|\]|T} $tspec " " tspec

    set r [ convert_2_k [lindex $tspec 0]]
    set b [ convert_2_k [lindex $tspec 1]]
    set p [ convert_2_k [lindex $tspec 2]]
    set m [ convert_2_k [lindex $tspec 3]]
    set M [ convert_2_k [lindex $tspec 5]]
    
    return [concat $r $b $p $m $M]

}
#
# send a message indicating the received event
#
proc Message  { host port mesg } {
    global sessions

    foreach i $sessions {
	if {[string match *$host:$port* $i]} {
	    pmm_send "$i message $mesg"
	}
    }
}


# Add source to the source list and inform  isc about
proc add_src {host port src tspec} {
    global pmm sessions path_msg tspec_msg MODE

    if {$pmm(debug) > 0} {
	puts stderr "host -> $host port -> $port    tspec -> $tspec"
    }

    foreach i $sessions {
	if {[string match *$host:$port* $i]} {
	    set ses $i
	    set path_msg [ladd  $path_msg "$ses add_src $src $host $port"]     
	    set tspec_msg \
		[ladd $tspec_msg "$ses flowspec $src [pmm_tspec $tspec]"]
	    
	    if { $MODE == "_sm" } {
		rapi_fspec_add_sm $host $port $src [pmm_tspec $tspec]
	    } else {
		pmm_send "$ses add_src $src $host $port"
		pmm_send "$ses flowspec $src \{[pmm_tspec $tspec]\}" 
	    }
	}
    }
}

# Remove source to the source list and inform about isc
proc rm_src {host port src} {
    global pmm sessions MODE
    
    foreach i $sessions {
	if {[string match "*$host:$port*" $i]} {
	    set ses $i
	    if { $MODE == "_sm" } {
		rapi_fspec_rm_sm $host $port $src
	    } else {
		pmm_send "$ses rm_src $src $host $port"
	    }
	}
    }
}

# initialize some global variables for the standalone mode
set map [list 0]
set id 0
set CONFIGURE_LEN 3

#
#find an empty slot for the rsvp entry
#
proc get_location {} {
    global map
    if { [lsearch -exact $map 0 ] < 0 } {
	set position [llength $map]
	set map [lappend map 1]
    } else {
	set position [lsearch -exact $map 0 ]
	set map [lreplace $map $position $position 1]
    }
    return $position
}


#
# Display status message at bottom of isc.
#
proc footer {msg} {
    global id

    if { $id != 0 } {
	after cancel $id
    }
    
    .message_win.message configure -text $msg
    set id [after 20000 "footer {}"]
}

#
# TSPEC-specific settings.
#
proc config_spec {{w .reservation} Entry spec {m ""}} {
    global Entries

    set w $w.$spec
    if [winfo exists $w] return
    toplevel $w
    set lw [string length "Max :  "] 
    
    if {$m != ""} {
	set index [set Entry]_[set m]
    } else {
	set index $Entry
    }
    
    frame $w.up
    pack \
	[tixControl $w.up.type -label "Type" -max 6 -min 0 -step 1 \
	     -variable  Entries($Entry,[set spec]_type) \
	     -options { label.width 12} ] \
	[tixControl $w.up.b -label "b (byte)" -max 1000000 -min 0 -step 1000\
	     -variable  Entries($index,[set spec]_b)  \
	     -options { 
		 label.width 12 
		 label.anchor e
		 entry.borderwidth 3
		 entry.width 10 }] \
	[tixControl $w.up.r -label "r (byte/s)" -max 1000000 -min 0 -step 10\
	     -variable  Entries($index,[set spec]_r)  \
	     -options { label.width 12 
		 label.anchor e
		 entry.borderwidth 3
		 entry.width 10 }] \
	[tixControl $w.up.p -label "p (byte/s)" -max 1000000 -min 0 -step 10\
	     -variable  Entries($index,[set spec]_p)  \
	     -options { label.width 12 
		 label.anchor e
		 entry.borderwidth 3
		 entry.width 10 }] \
	[tixControl $w.up.m -label "m (byte)" -max 1000000 -min 0 -step 1000\
	     -variable  Entries($index,[set spec]_m)  \
	     -options { label.width 12 
		 label.anchor e
		 entry.borderwidth 3
		 entry.width 10 }] \
	[tixControl $w.up.m2 -label "M (byte)" -max 1000000 -min 0 -step 1000\
	     -variable  Entries($index,[set spec]_M)  \
	     -options { label.width 12 
		 label.anchor e
		 entry.borderwidth 3
		 entry.width 10 }] \
    -side top -anchor w -expand yes -fill both 

	pack $w.up -side top
  frame $w.action
  pack \
    [button $w.action.accept  -text "Dismiss"  -relief raised\
	 -command "destroy $w"] \
    [button $w.action.apply  -text "Apply"  -relief raised\
	 -command "destroy $w;rapi_[set spec]_sm $Entry $m"] \
     -side left -expand yes -fill both
  pack $w.action -expand yes -fill both
}
#
# enable the button of an entry
#
proc enable_entry {Entry name {m ""}} {
    global Entries
    
    set member_win ""
    set index $Entry

    set des $Entries($Entry,destination)
    set w .configure_win
    if { $m!="" } {
	regsub -all {\.} $m  _ member_win 
	set member_win _$member_win
	set index [set Entry]_[set m]
	set des $m
    }	
    

    $Entries($Entry,f_win).[set name].receiver$member_win.fspec configure \
	-state normal -fg black 
    $Entries($Entry,f_win).[set name].receiver$member_win.reserve configure \
	-state normal -fg black
}

#
# disable the button of an entry
#
proc disable_entry {Entry name {m ""}} {
    global Entries
    
    set member_win ""
    set index $Entry

    set des $Entries($Entry,destination)
    set w .configure_win
    if { $m!="" } {
	regsub -all {\.} $m  _ member_win 
	set member_win _$member_win
	set index [set Entry]_[set m]
	set des $m
    }	
    
    $Entries($Entry,f_win).[set name].receiver$member_win.fspec configure \
	-state disabled -fg grey 
    $Entries($Entry,f_win).[set name].receiver$member_win.reserve configure \
	-state disabled -fg grey
}

#
# disable/enable the appropriate members for making reservations
# based on the filter type
#
proc update_filter {name element op} {
    global Entries


    set Entry [lindex [split $element ,] 0]
    if { [llength $Entries($Entry,members)] == 0} return
    if {$Entries($Entry,filter) == "Wild Card" } {
	rsvp rapi_filtertype $Entries($Entry,destination) \
	    $Entries($Entry,port)  0
	enable_entry $Entry session
	for {set i 0} {$i<[llength $Entries($Entry,members)]} {incr i} {
	    set member [lindex $Entries($Entry,members) $i]
	    disable_entry $Entry members $member
	}
	
    }

    if { $Entries($Entry,filter) == "Fixed" } {
	rsvp rapi_filtertype $Entries($Entry,destination) \
	    $Entries($Entry,port)  1
	for {set i 0} {$i<[llength $Entries($Entry,members)]} {incr i} {
	    set member [lindex $Entries($Entry,members) $i]
	    enable_entry $Entry members $member
	}
	disable_entry $Entry session
    }	    
    
}

#
# build the entry for each member
#
proc open_entry {Entry name {m ""}} {
    global Entries
    
    set member_win ""
    set index $Entry

    set des $Entries($Entry,destination)
    set w .configure_win
    if { $m!="" } {
	regsub -all {\.} $m  _ member_win 
	set member_win _$member_win
	set index [set Entry]_[set m]
	set des $m
    }	
    
    set win [frame $Entries($Entry,f_win).[set name].receiver$member_win ]
    pack $win -side top -anchor w
    pack \
	[button $win.fspec \
	     -text "Flow Spec" \
	     -command "config_spec $w.$Entry $Entry fspec $m " \
	     -state disabled -width 15 -fg grey]\
	[checkbutton $win.reserve \
	     -text "Reserve: $des :$Entries($Entry,port)" -state disabled\
	     -fg grey -variable Entries($index,reserve) \
	     -command "rapi_reserve_sm $Entry $m"]\
	-side left  -anchor w
}

#
# open the configuration window at row,column for Entry
#
proc open_configure {Entry win} {
    global Entries CONFIGURE_LEN

    set w .configure_win
    if { [winfo exists $w.$Entry] } return
    set position [get_location]
    set row [expr $position / $CONFIGURE_LEN]
    set column [expr $position % $CONFIGURE_LEN]

    set Entries($Entry,config_row) $row
    set Entries($Entry,config_column) $column
    

    grid configure [frame $w.$Entry \
			-relief groove -borderwidth 1]\
	-row $row -column $column -sticky news
    tixLabelFrame $w.$Entry.border \
	-label $Entries($Entry,destination):$Entries($Entry,port)
    set border_f [$w.$Entry.border subwidget frame]

    # path control stuff
    tixLabelFrame $border_f.path -label "PATH" 
    set f_path [$border_f.path subwidget frame]

    pack \
	[button $f_path.tspec -text "Sender Tspec" \
	     -command "config_spec $w.$Entry $Entry tspec"]\
	[checkbutton $f_path.path \
	     -text "PATH:$Entries($Entry,destination):$Entries($Entry,port)" \
	     -relief flat -variable Entries($Entry,path)\
	     -command "rapi_path_sm $Entry"] \
	-side left -expand yes -fill both

 	
    # reserve control stuff
    tixLabelFrame $border_f.reserve -label "RESERVE" 
    set f_reserve [$border_f.reserve subwidget frame]
    set Entries($Entry,f_win) $f_reserve

    # frame for the wild card reservation stuff
    set f_reserve [frame $Entries($Entry,f_win).session ]
    open_entry $Entry session 
  

    # frame for the fixed filter reservation stuff
    set f_reserve [frame $Entries($Entry,f_win).members  ]

    # chech if there are any members already in the list and if so build their 
    # corresponding widgets
    if { [llength $Entries($Entry,members)] >0} {
	for {set i 0} {$i<[llength $Entries($Entry,members)]} {incr i} {
	    open_entry $Entry members [lindex $Entries($Entry,members) $i]
	}
    }

    
    # frame for the filter entry
    set f_reserve [frame $Entries($Entry,f_win).filter  ]
    pack \
	[tixOptionMenu $f_reserve.filter_type -variable Entries($Entry,filter)\
	     -width 15 -state disabled -options {
		 menubutton.width 10
	         menubutton.relief raised }] \
	-side top
    trace variable Entries($Entry,filter) w update_filter
    if { [llength $Entries($Entry,members)] >0} {
	$f_reserve.filter_type configure -state normal

	# while we are here we can make sure that the appropriate
	# member widgets are enabled by setting the filter variable
	set Entries($Entry,filter) $Entries($Entry,filter) 
    }

    $f_reserve.filter_type add command "Wild Card"
    $f_reserve.filter_type add command "Fixed"
    
    pack $Entries($Entry,f_win).session $Entries($Entry,f_win).members \
	$Entries($Entry,f_win).filter \
	-side top -expand yes -fill both

    pack $border_f.path $border_f.reserve -side top -expand yes -fill both

    pack $w.$Entry.border \
	[button $border_f.dismiss -relief raised -text "Dismiss"\
	     -padx 4 -command "destroy $w.$Entry; destroy_configure $Entry"]\
	[button $border_f.close -relief raised -text "End RSVP"\
	     -padx 4 -command "destroy $w.$Entry;\
                   rapi_release_sm $Entry $win"]\
	-side left  -anchor w -fill both -expand yes
    
}
#
# destroy the configuration window for Entry
#
proc destroy_configure { Entry} {
    global Entries map CONFIGURE_LEN
    
    # release the slot in the map
    set index [ expr [expr $Entries($Entry,config_row) * $CONFIGURE_LEN] \
		    + $Entries($Entry,config_column)]
    set map [lreplace  $map  $index $index  0]
    trace vdelete Entries($Entry,filter) w update_filter
}  

#
# configure the reservation part of the configure win for Entry
#
proc enable_reservation_win {Entry m} {
    global Entries

    # add a possible new member to the member list
    set index [set Entry]_[set m]
    if { [lsearch -exact $Entries($Entry,members)  $m] < 0 } {
	set  Entries($Entry,members) [lappend Entries($Entry,members) $m]
	set  Entries($index,reserve) 0
    }
     set w .configure_win
    if { ![winfo exists $w.$Entry] } return
    
    if { $Entries($Entry,filter) == "Wild Card" } {
	enable_entry $Entry session
    }
    # build the actual widgest
    open_entry $Entry members $m

    $Entries($Entry,f_win).filter.filter_type configure -state normal 

    if { $Entries($Entry,filter) == "Fixed" } {
	for {set i 0} {$i<[llength $Entries($Entry,members)]} {incr i} {
	    set member [lindex $Entries($Entry,members) $i]
	    enable_entry $Entry members $member
	}
    }	    
}

#
# rm the flow spec for member m
#
proc rapi_fspec_rm_sm {destination port m} {
    global Entries
    # replace all . with _ to regenerate the Entry
    regsub -all {\.} $destination _ tmp
    set Entry "$tmp:$port"

    set index [set Entry]_[set m]
    if { [lsearch -exact $Entries($Entry,members) $m] >-1 } { 
	# remove the parameters for index
	unset Entries($index,fspec_type) 
	unset Entries($index,fspec_b) 
	unset Entries($index,fspec_r) 
	unset Entries($index,fspec_p)
	unset Entries($index,fspec_m) 
	unset Entries($index,fspec_M) 
	
	# remove the entry in members
	set l [lsearch -exact $Entries($Entry,members) $m]
	set  Entries($Entry,members) [lreplace $Entries($Entry,members) $l $l]
	
	# close the corresponding widgets
	set w .configure_win

	regsub -all {\.} $m  _ member_win 
	if { ![winfo exists $w.$Entry] } return
	destroy $Entries($Entry,f_win).members.receiver_$member_win
    }
    if { [llength $Entries($Entry,members)] == 0 } {
	set Entries($Entry,fspec_type) 5 
	set Entries($Entry,fspec_b) 0
	set Entries($Entry,fspec_r)  0
	set Entries($Entry,fspec_p) 0
	set Entries($Entry,fspec_m) 0
	set Entries($Entry,fspec_M) 0

	set Entries($Entry,reserve) 0

	disable_entry $Entry session
	$Entries($Entry,f_win).filter.filter_type configure -state disabled 
    }
}	 
 
#
# update the flow spec for member m
#
proc rapi_fspec_add_sm {destination port m flowspec} {
    global Entries
    
    set flowspec [split $flowspec " "]
    # replace all . with _ to regenerate the Entry
    regsub -all {\.} $destination _ tmp
    set Entry "$tmp:$port"

    set index [set Entry]_[set m]

    set Entries($index,fspec_type) 5
    set Entries($index,fspec_b) [expr 1000 * [lindex $flowspec 1]]
    set Entries($index,fspec_r) [lindex $flowspec 0]
    set Entries($index,fspec_p) [lindex $flowspec 2]
    set Entries($index,fspec_m) [lindex $flowspec 3]
    set Entries($index,fspec_M) [lindex $flowspec 4]

    # set the values for the flowspec to the biggest seen flowspec values
    # from all receivers
    foreach i "b r p m M" {
	if {$Entries($Entry,fspec_$i) < $Entries($index,fspec_$i)} {
	    set Entries($Entry,fspec_$i)  $Entries($index,fspec_$i)
	}
    }
    set Entries($Entry,fspec_type) 5

    enable_reservation_win $Entry $m
    
}

#
# update the fspec of Entry
#
proc rapi_fspec_sm {Entry {m ""}} {
    global Entries
    
    
    if {$m != ""} {
	set index [set Entry]_[set m]
    } else {
	set index $Entry
    }

    #update the tspec value
    rsvp rapi_flowspec $Entries($Entry,destination) $Entries($Entry,port) \
	$Entries($index,fspec_type) \
	$Entries($index,fspec_b) \
	$Entries($index,fspec_r) \
	$Entries($index,fspec_p) \
	$Entries($index,fspec_m) \
	$Entries($index,fspec_M) \
	0 0 $m

    #force a new reserve message to be sent with the new fspec
    if { $Entries($index,reserve) >0 } {
	rapi_reserve_sm $Entry $m
    }
}
#
# update the tspec of Entry
#
proc rapi_tspec_sm {Entry {m ""}} {
    global Entries
    
    #update the tspec value
    rsvp rapi_tspec $Entries($Entry,destination) $Entries($Entry,port) \
	$Entries($Entry,tspec_type) \
	$Entries($Entry,tspec_b) \
	$Entries($Entry,tspec_r) \
	$Entries($Entry,tspec_p) \
	$Entries($Entry,tspec_m) \
	$Entries($Entry,tspec_M)

    #force a new path message to be sent with the new tspec
    if { $Entries($Entry,path) >0 } {
	rapi_path_sm $Entry
    }
}

#
# send or stop sending path messages for the sm state
#
proc rapi_path_sm {Entry} {
    global Entries

    rsvp rapi_sender $Entries($Entry,destination) $Entries($Entry,port) \
	$Entries($Entry,path)
}

#
# send or stop sending path messages for the sm state
#
proc rapi_reserve_sm {Entry {m ""}} {
    global Entries

    if { $m =="" } {
	rsvp reserve $Entries($Entry,destination) $Entries($Entry,port) \
	    $Entries($Entry,reserve)
    } else {
	set index [set Entry]_[set m]
	rsvp set_reservation  $Entries($Entry,destination) \
	    $Entries($Entry,port) \
	    $Entries($index,reserve) \
	    $m
	rsvp reserve  $Entries($Entry,destination) \
	    $Entries($Entry,port) \
	    $Entries($index,reserve) \
	    $m
    }
	
}

#
# end rsvp session
#
proc rapi_release_sm {Entry w} {
    global Entries map CONFIGURE_LEN sessions

    # release the rsvp session
    rsvp rapi_release $Entries($Entry,destination) $Entries($Entry,port)
    
    # delete the Entry from the sessions list
    set sessions \
	[lremove $sessions $Entries($Entry,destination):$Entries($Entry,port)]

    # release the slot in the map
    set index [ expr [expr $Entries($Entry,config_row) * $CONFIGURE_LEN] \
		    + $Entries($Entry,config_column)]
    set map [lreplace  $map  $index $index  0]

    set l [llength [array names Entries]]
    destroy $w.label_$Entry
    set m 0
    for { set i 0} {$i< $l} { incr i } {
	if { [lsearch -exact [lindex [split [lindex [array names Entries] $m] \
					  ","]  0] $Entry] >-1} {
	    unset Entries([lindex [array names Entries] $m])
	} else {
	    incr m
	}
	
    }

}

#
# open the rsvp session
#
proc rapi_session_sm { w } {
    global destination_sm host_sm  port_sm Entries sessions
    

    if {$destination_sm == "" } {
	footer "No destination specified"
	return 0
    }
    if {$port_sm == "" } {
	footer "No port specified"
	return 0
    }
    
    # replace all . with _
    regsub -all {\.} $destination_sm _ tmp
    set Entry "$tmp:$port_sm"

     # add the session name to the list 
    if { [lsearch -exact [split [array names Entries] ","] $Entry] < 0 } {
	
	#add to the sessions list
	set sessions [ladd $sessions $destination_sm:$port_sm]

	# open the RSVP session 
	rsvp rapi_session $destination_sm $port_sm  
	rsvp set_local $destination_sm $port_sm $host_sm

	# init the entry strycture 
	set Entries($Entry,destination) $destination_sm
	set Entries($Entry,port) $port_sm
	set Entries($Entry,row) \
	    [lindex [grid size $w] 1]
	
	#init the tspec
	set Entries($Entry,tspec_type) 5
	set Entries($Entry,tspec_b) 16000
	set Entries($Entry,tspec_r) 64
	set Entries($Entry,tspec_p) 0
	set Entries($Entry,tspec_m) 0
	set Entries($Entry,tspec_M) 65535

	#init the fspec
	set Entries($Entry,fspec_type) 5
	set Entries($Entry,fspec_b) 0
	set Entries($Entry,fspec_r) 0
	set Entries($Entry,fspec_p) 0
	set Entries($Entry,fspec_m) 0
	set Entries($Entry,fspec_M) 0

	#init the filter type
	set Entries($Entry,filter) "Wild Card"

	#init the member count
	set Entries($Entry,members) ""

	# add an entry for this session in the session list
	pack [label $w.label_$Entry \
		  -text $Entries($Entry,destination):$Entries($Entry,port) \
		  -bg LemonChiffon1 -cursor gumby] \
	    -side top -anchor w
	bind $w.label_$Entry <ButtonPress-1> "open_configure $Entry $w"
	open_configure $Entry $w
    } else {
	footer "Session $Entry already opened"
	return
    }	

    
}

#
# build the main window for the isc controlled version
#
proc build.window w {
    wm geometry $w 100x100+0+0
    wm title $w "PMM_RSVP:[idstring %u@%d]"
    update idletasks
    wm iconify $w
    pack [button .quit -text "Quit!" -command adios_exit ]

    update idletasks
}


#
# build the window for the stand alone modul
#
proc build.window_sm w {
    global Entries

    wm title $w "PMM_RSVP:[idstring %u@%d]"
    update idletasks


    #initialize variables
    global destination_sm host_sm  port_sm
    set host_sm [idstring "%h"]
    set destination_sm ""
    set port_sm ""

    # sessions list window
    tixScrolledWindow  .list_win -scrollbar both \
	 -height 200 -relief groove -borderwidth 1
    # create the conatiner window for the the members
    set list_frame [frame [.list_win subwidget window].frame -bg LemonChiffon1]
    pack $list_frame -anchor w 
    
    #Session addition frame
    set add_frame [frame .add_win -relief groove -borderwidth 2]
    pack \
	[ button $add_frame.add -relief raised -text "Add Session" \
	      -command "rapi_session_sm $list_frame"]\
	[label $add_frame.des_label -text Destination]\
	[entry $add_frame.des_entry -textvariable destination_sm \
	     -background white ]\
	[label $add_frame.host_label -text Host]\
	[entry $add_frame.host_entry -textvariable host_sm \
	     -background white]\
	[label $add_frame.port_label -text Port]\
	[entry $add_frame.port_entry -textvariable port_sm \
	     -background white ]\
	-side left  -anchor w



    # frame for the configuration windows
    set configure_frame [frame .configure_win -relief groove -borderwidth 2]
    

    #message window
    set message_frame [frame .message_win -relief groove \
		       -borderwidth 2 -bg white]
    message $message_frame.message -aspect 10000 -anchor w -bg white -fg blue
    pack $message_frame.message

    pack $add_frame .list_win $list_frame $configure_frame $message_frame \
	[button .quit -text "Quit!" -command adios_exit]\
	-side top  -anchor w -expand yes -fill both
    update idletasks
}

