#
# pmm.tcl - this file implements the pmm protcol for vic
#
# Initial Version written by K. Samp 
# Modified and enhanced version written by F. Emanuel
#
# $Id: cf-pmm.tcl,v 1.2 1998/07/20 22:39:39 koepsel Exp $
# 
# ----------------------------------------------------------------------

#
# Initialisation of pmm - invoke pmm_module and open_menu
# after vic_main has started

set pmm(tool) ""
set c ""
set myself ""
set first 1
set srclist {}

#
# Called from the pmm_module function.
#
proc init_pmm {} {
    global pmm  
    

#    rename adios pmm_adios
#    proc adios {} {
#	global pmm myself 
#	set src [srctab local]
#	set src $pmm($myself)
#	set cname [$src sdes cname]
#	if [info exists pmm(sn)] {
#	    pmm_send_ " closed"
#	}
#	puts "\n$pmm(tool): Exiting...\n"
#	pmm_adios
#    }

    set pmm(debug) 1
    set pmm(controller) "225.2.0.1 5010"
    set pmm(tool) "vat"
    
    pmm_init_udp $pmm(controller)


    # Issue "attached" message
    pmm_send "agent $pmm(tool) attached [pid] audio"

}



#
# Open UDP connection 
#
proc pmm_init_udp {controller} {
   global pmm c
    
    set host [lindex $controller 0]
    set port [lindex $controller 1]
    set c [udp $host $port]
    #Tcl_SetSockBuffer $c
    fconfigure $c -ttl 0  
    fconfigure $c -buffering none
    fconfigure $c -member 1
    fconfigure $c -buffersize 128000
    fileevent $c readable {pmm_read_udp $c}
}    

#
# Receive UDP packet
#
proc pmm_read_udp {sock} {
    pmm_eval [gets $sock]
}

#
# Source and destination port on all multicast packets is 5010
# (sending port uses bind(), so that we have no way of knowing whether
# we see our own transmissions or those from another process.  
#
proc pmm_eval {cmd} {
    global pmm myself meterDisable mute cnamelist

    set result [catch {exec id} id]
    set id2 $id
    set user [string trim [string range $id2 [string first ( $id] [string first ) $id]] "()"] 

    set user [exec whoami]
    set hostlist [exec hostname -i]
    set host [lindex $hostlist 0]
    #set host [exec hostname -f]
    set myself "$user@$host"
    #puts stderr $myself


    if {[string length $cmd] > 0} {
	#puts stderr "PMM -> $cmd"
	if {[lsearch {proc exec rename} [lindex $cmd 0]] >= 0} {
	    puts stderr "[pid] $pmm(tool): forbidden command: $cmd"
	} elseif {[string match "session *" $cmd]} {
	    # Set to generic pattern on first invocation, but then only
	    # allow repeat of first 'session' command.  Must not answer
	    # other video session requests.
	    #puts stderr "session found!"
	    if {![info exists pmm(sn)]} {

		set textvar1 $cmd
		set end1 [string first / $textvar1]
		incr end1 -1
		set sessiontitle [string trim [string range $textvar1 [string first " " $textvar1] $end1]]
		#puts stderr "sessiontitle = $sessiontitle"
		set pmm(sn) $sessiontitle/audio/*
		#puts stderr "pmm(sn) $pmm(sn)"
	    }
	    if {[string match $pmm(sn) [lindex $cmd 1]]} {
		set pmm(sn) [lindex $cmd 1]
		pmm_send_ " created"
		pmm_send_ " hello $myself"
		wm geometry . 90x0+0+0
		wm iconify .

		#puts stderr "myself=$myself"
		#set cnamelist($myself) $myself
		set mute($myself) 0
	    }
	} else {
	    if {[info exists pmm(sn)]&&[string match $pmm(sn)* $cmd]} {
		#puts stderr "$pmm(sn) $cmd"
		pmm_eval_ $cmd 
	    }
	}
    }
}


#
# Send PMM message
#
proc pmm_send {msg} {
    global pmm c
    
    puts -nonewline $c $msg 
}


#
# Like pmm_send, but prefix our session name.
#
proc pmm_send_ {msg} {
    global pmm
    
    if [info exists pmm(sn)] {
	pmm_send "$pmm(sn)$msg"
    }
}

#
# ignore pmm "+" message
#
proc pmm_+ {args} {
}
 
#
# ignore pmm "agent" message
#
proc pmm_agent {args} {
}

proc send_update {} {
    global pmm

    pmm_send_ " active $pmm(update_interval)"
    after $pmm(update_interval) send_update

}



#
# Invoked when a pmm session command was received.
# Vic-specific since it invokes buttons.
#
proc pmm_eval_ {msg} {
    global pmm V first mute srclist cnamelist myself stat_method

    set c [lindex $msg 1]
    string trim $c
    switch $c {
	hello {
	}
	update_interval {
            set pmm(update_interval) [lindex $msg 2]
            send_update
        }
	cname {
	}
	open {
	    set net $V(data-net)
	    pmm_send_ " opened [$net addr]"
	}
	opened {
	}
	close {
	    pmm_send_ " closed {}"
	    adios
	}
	closed {
	}
	media {
	    set type [lindex $msg 2]
	    switch $type {
		"PCMU 1 8000" {
		    select_format pcm 2
		}
		"DVI4 1 8000" {
		    select_format dvi 4
		}
	        "L16 1 44100" {

		}
		"LPC 1 8000" {
		    select_format lpc 4
		}
	    }
	}
	transport {
	}
	send {
	    set type [lindex $msg 2]
	    #puts stderr "SEND"
	    switch $type {
		1 {
		    set_recvonly_mode 0
		    if ![audio have] {
			audio_demand
		    }
		    foreach elem $srclist {
			#puts stderr "src --> $elem"
			set v [expr ![$elem mute]]
			#puts stderr "toggle_mute: v -> $v"
			
			if { "$v" == "1" } {
			    toggle_mute $elem
			    pmm_send_ "/$pmm($elem) receiving 0"
			    #puts stderr "Source $elem abschalten!!"
			}
		    }
		    audio input unmute
		}
		0 {
		    audio input mute	
		    set_recvonly_mode 1
		    foreach elem $srclist {
			set v [expr ![$elem mute]]
			if { "$v" == "0" } {
			    toggle_mute $elem
			    pmm_send_ "/$pmm($elem) receiving 1"
			    #puts stderr "Source $elem wieder einschalten!!"
			}
		    }
		}
	    }

	}
	receive {

	    set textvar2 [lindex $msg 0]
	    #puts stderr "receive -> $textvar2"	    
	    set start1 [string last "/" $textvar2]
	    incr start1
	    set cname [string trim [string range $textvar2 $start1 [string length $textvar2]]] 
	    #puts stderr "RECEIVE -> cname=$cname"
	    if { "$mute($cname)" == "1" } {
		toggle_mute $pmm($cname)
	    }
	    set mute($cname) 1
	}
	statistics {
	}
	statistics? {
	    set textvar2 [lindex $msg 0]
	    set start1 [string last "/" $textvar2]
	    incr start1
	    set cname [string trim [string range $textvar2 $start1 [string length $textvar2]]] 
	    set w $pmm($cname)
	    set stats [eval $w stats_pmm]
	    #puts stderr "stats -> $stats"
	    pmm_send_ "/$cname statistics $stats"
	}

    }
}


#
# This function returns the dotted notation of another host
# by resolving the given name from a hello cmd
# e.g.: user@host.domain.net -> 192.168.0.1
#       user@192.168.0.1     -> 192.168.0.1
#
proc get_host_from_cname {user_host} {
    set  addrstr $user_host
    set  start  [string first "@" $addrstr]
    incr start
    set  end    [string length $addrstr]
    set  addr   [string trim [string range $addrstr $start $end]]

    set  dotted [getdottedaddr $addr]

    return $dotted

}

#
# This function returns the user from a cname string
# received by a hello cmd
#
proc get_user_from_cname {user_host} {
    set  userstr $user_host
    set  end  [string first "@" $userstr]
    incr end -1
    set  user   [string trim [string range $userstr 0 $end]]

    return $user

}                              

proc register_pmm src {
    global pmm mute first srclist cnamelist myself
	#
	# if we display everyone (i.e., non-mixers and mixers),
	# then just create the name in the sitebox.  otherwise,
	# it gets deferred until we receive an SDES packet
	# (which never happens for a mixer)
	#
        #puts stderr "register_pmm $src called"
        set cname [$src sdes cname]
        if {![info exists pmm($cname)]} {    
	    if { "$first" != "1" } {
		lappend srclist "$src"
	    }
	    set pmm($cname) "$src"
	    set pmm($src) "$cname"
	}
	#puts stderr "CNAME -> $cname ---- pmm(cname) -> $pmm($cname)"
        if { "$cname" != "" } {
	    #puts stderr "cname -> $cname"
	    #set host_ip [get_host_from_cname $cname]
	    #set user    [get_user_from_cname $cname]
	    #set isc_cname "$user@$host_ip"
	    #puts stderr "ISC -> cname=$isc_cname"

	    #set cnamelist($cname) $isc_cname
	    #set cnamelist($isc_cname) $cname

	    pmm_send_ " hello $cname"
	    #puts stderr "setting mute to 0 for $cname"
	    set mute($cname) 0


	    if { "$first" == "1" } {
		# Hide the vat main window simply by setting its geoemtry to 90x0 :)
		set pmm(local) "$cname"
		set pmm(localsrc) "$src"
		set first 0 
	    }
	    #puts stderr "local src -> $pmm(localsrc)"

        }
}

proc unregister_pmm src {
    global pmm srclist cnamelist
        #puts stderr "unregister_pmm $src"
        set cname [$src sdes cname]
	#puts stderr "CNAME -> $cname"
        if { "$cname" != "" } {
	    pmm_send_ "/$cname bye"
	    #puts stderr "-- $src --"
	    set remove_src [lsearch $srclist $src]
	    #puts stderr "remove_src -> $remove_src"
	    if { $remove_src != -1 } {
		set srclist_ [lreplace $srclist $remove_src $remove_src]
		set srclist $srclist_
	    }

        }
}


proc set_busy_pmm src {
    global pmm src_name myself cnamelist

    

    #puts stderr "local src -> $pmm(localsrc)"
    set cname [$src sdes cname]
    if {[info exists src_name($src)]} {
	#puts stderr "-- $cname --"
	if {$cname != ""} {
		if { $pmm(localsrc) == $src } {
			pmm_send_ "/$myself active"
		} else {
	    		pmm_send_ "/$cname active"
		}
	}
    }

}

proc clear_busy_pmm src {
    global pmm src_name myself cnamelist

    #puts stderr "clear_busy_pmm"
    set cname [$src sdes cname]
    if [info exists src_name($src)] {
	if {$cname != ""} {
            if { $pmm(localsrc) == $src } {
		pmm_send_ "/$myself !active"
	    } else {
	    	pmm_send_ "/$cname !active"
            }
	}
    }

}

