#include <sys/types.h>
#include <netinet/in.h>  /* inet_ntoa() */
#include <arpa/inet.h>
#include <malloc.h>      /* free() */
#include <tcl.h>         /* from Tcl/tk package */
#include "servers.h"


/*
* Tcl routine that returns a list of the form
*  {{host ip-address preference} {...}} for each SIP candidate server.
*/
int servers_cmd(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  int i;
  server_t *s;

  if (argc != 2) {
    sprintf(interp->result, "Usage: %s domain", argv[0]);
    return TCL_ERROR;
  }
  s = servers(argv[1]);

  /* Sort and print list and remove each entry when done. */
  for (i = 0; s[i].host; i++) {
    char temp[1024];
    sprintf(temp, "%s %s %d",
      s[i].host, inet_ntoa(s[i].addr), s[i].preference);
    Tcl_AppendElement(interp, temp);
    free(s[i].host);
  }
  if (s) free(s);
  return TCL_OK;
} /* servers_cmd */


/*
* Install as Tcl command.
*/
int Servers_Init(Tcl_Interp *interp)
{
  Tcl_CreateCommand(interp, "servers", servers_cmd, (ClientData)NULL,
    (Tcl_CmdDeleteProc *)0);
  return TCL_OK;
} /* Servers_Init */
